'''
Library for specular and off-specular x-ray reflectivity
========================================================
interdiff is a model for specular and off specular simulations including
the effects of interdiffusion in hte calculations. The specular
simulations is conducted with Parrats recursion formula. The
off-specular, diffuse calculations are done with the distorted Born wave
approximation (DWBA) as derived by Holy and with the extensions done by
Wormington to include diffuse interfaces.

Classes
-------

Layer
~~~~~
``Layer(b = 0.0, d = 0.0, f = 0.0+0.0J, dens = 1.0, magn_ang = 0.0, magn = 0.0, sigma = 0.0)``

``d``
   The thickness of the layer in AA (Angstroms = 1e-10m)
``f``
   The x-ray scattering length per formula unit in electrons. To be
   strict it is the number of Thompson scattering lengths for each
   formula unit.
``dens``
   The density of formula units in units per Angstroms. Note the units!
``sigmai``
   The root mean square *interdiffusion* of the top interface of the
   layer in Angstroms.
``sigmar``
   The root mean square *roughness* of the top interface of the layer in
   Angstroms.

Stack
~~~~~
``Stack(Layers = [], Repetitions = 1)``

``Layers``
   A ``list`` consiting of ``Layer``\ s in the stack the first item is
   the layer closest to the bottom
``Repetitions``
   The number of repsetions of the stack

Sample
~~~~~~
``Sample(Stacks = [], Ambient = Layer(), Substrate = Layer(), eta_z = 10.0,     eta_x = 10.0, h = 1.0)``

``Stacks``
   A ``list`` consiting of ``Stack``\ s in the stacks the first item is
   the layer closest to the bottom
``Ambient``
   A ``Layer`` describing the Ambient (enviroment above the sample).
   Only the scattering lengths and density of the layer is used.
``Substrate``
   A ``Layer`` describing the substrate (enviroment below the sample).
   Only the scattering lengths, density and roughness of the layer is
   used.
``eta_z``
   The out-of plane (vertical) correlation length of the roughness in
   the sample. Given in AA.
``eta_x``
   The in-plane global correlation length (it is assumed equal for all
   layers). Given in AA.
``h``
   The jaggedness parameter, should be between 0 and 1.0. This describes
   how jagged the interfaces are. This is also a global parameter for
   all interfaces.

Instrument
~~~~~~~~~~
``Instrument(wavelength = 1.54, coords = 'tth',      I0 = 1.0 res = 0.001, restype = 'no conv', respoints = 5, resintrange = 2,      beamw = 0.01, footype = 'no corr', samplelen = 10.0, taylor_n = 1)``

``wavelength``
    The wavalelngth of the radiation givenin AA (Angstroms)
``coords``
    The coordinates of the data given to the SimSpecular function. The
    available alternatives are: 'q' or 'tth'. Alternatively the numbers 0
    (q) or 1 (tth) can be used.
``I0``
    The incident intensity (a scaling factor)
``Ibkg``
    The background intensity. Added as a constant value to the calculated
    reflectivity
``res``
    The resolution of the instrument given in the coordinates of ``coords``.
    This assumes a gaussian reloution function and ``res`` is the standard
    deviation of that gaussian.
``restype``
    Describes the rype of the resolution calculated. One of the
    alterantives: 'no conv', 'fast conv', 'full conv and varying res.' or
    'fast conv + varying res.'. The respective numbers 0-3 also works. Note
    that fast convolution only alllows a single value into res wheras the
    other can also take an array with the same length as the x-data (varying
    resolution)
``respoints``
    The number of points to include in the resolution calculation. This is
    only used for 'full conv and vaying res.' and 'fast conv + varying res'
``resintrange``
    Number of standard deviatons to integrate the resolution fucntion times
    the relfectivty over
``footype``
    Which type of footprint correction is to be applied to the simulation.
    One of: 'no corr', 'gauss beam' or 'square beam'. Alternatively, the
    number 0-2 are also valid. The different choices are self explanatory.
``beamw``
    The width of the beam given in mm. For 'gauss beam' it should be the
    standard deviation. For 'square beam' it is the full width of the beam.
``samplelen``
    The length of the sample given in mm
``taylor_n``
    The number terms taken into account in the taylor expansion of the
    fourier integral of the correlation function. More terms more accurate
    calculation but also much slower.
'''
from .lib import paratt as Paratt
from .lib import offspec

from .lib.instrument import *

# Preamble to define the parameters needed for the models outlined below:
ModelID='MingInterdiff'
# Automatic loading of parameters possible by including this list
__pars__=['Layer', 'Stack', 'Sample', 'Instrument']
# Used for making choices in the GUI
instrument_string_choices={'coords': ['q', 'tth'],
                           'restype': ['no conv', 'fast conv',
                                       'full conv and varying res.', 'fast conv + varying res.'],
                           'footype': ['no corr', 'gauss beam', 'square beam']}

InstrumentParameters={'wavelength': 1.54, 'coords': 'tth', 'I0': 1.0, 'res': 0.001,
                      'restype': 'no conv', 'respoints': 5, 'resintrange': 2.0, 'beamw': 0.01, 'footype': 'no corr',
                      'samplelen': 10.0, 'Ibkg': 0.0, 'taylor_n': 1}
# Coordinates=1 => twothetainput
# Coordinates=0 => Q input
# Res stddev of resolution
# ResType 0: No resolution convlution
#               1: Fast convolution
#               2: Full Convolution +varying resolution
#               3: Fast convolution varying resolution
# ResPoints Number of points for the convolution only valid for ResolutionType=2
# ResIntrange Number of standard deviatons to integrate over default 2
# Parameters for footprint coorections
# Footype: 0: No corections for footprint
#          1: Correction for Gaussian beam => Beaw given in mm and stddev
#          2: Correction for square profile => Beaw given in full width mm
# Samlen= Samplelength in mm.

LayerParameters={'sigmai': 0.0, 'sigmar': 0.0, 'dens': 1.0, 'd': 0.0,
                 'f': 0.0+0.0j}
StackParameters={'Layers': [], 'Repetitions': 1}
SampleParameters={'Stacks': [], 'Ambient': None, 'Substrate': None, 'h': 1.0,
                  'eta_z': 10.0, 'eta_x': 10.0}

__xlabel__ = "q [Å$^{-1}$]"
__ylabel__ = "Instnsity [a.u.]"

def Specular(TwoThetaQz, sample, instrument):
    ''' Simulate the specular signal from sample when proped with instrument
    
    # BEGIN Parameters
    TwoThetaQz data.x
    # END Parameters
    '''
    # preamble to get it working with my class interface
    restype=instrument.getRestype()
    global __xlabel__
    __xlabel__ = "q [Å$^{-1}$]"

    if restype==2 or restype==instrument_string_choices['restype'][2]:
        (TwoThetaQz, weight)=ResolutionVector(TwoThetaQz[:],
                                              instrument.getRes(), instrument.getRespoints(),
                                              range=instrument.getResintrange())
    if instrument.getCoords()==1 or \
            instrument.getCoords()==instrument_string_choices['coords'][1]:
        theta=TwoThetaQz/2
        __xlabel__ = "2θ [°]"
    elif instrument.getCoords()==0 or \
            instrument.getCoords()==instrument_string_choices['coords'][0]:
        theta=arcsin(TwoThetaQz/4/pi*instrument.getWavelength())*180./pi

    lamda=instrument.getWavelength()
    parameters=sample.resolveLayerParameters()
    dens=array(parameters['dens'], dtype=float64)
    # print [type(f) for f in parameters['f']]
    f=array(parameters['f'], dtype=complex128)
    re=2.82e-13*1e2/1e-10
    n=1-dens*re*lamda**2/2/pi*f*1e-4
    d=array(parameters['d'], dtype=float64)
    # d = d[1:-1]
    sigmar=array(parameters['sigmar'], dtype=float64)
    # sigmar = sigmar[:-1]
    sigmai=array(parameters['sigmai'], dtype=float64)
    # sigmai = sigmai[:-1]
    sigma=sqrt(sigmai**2+sigmar**2)
    # print sigma

    R=Paratt.Refl(theta, lamda, n, d, sigma)*instrument.getI0()

    # FootprintCorrections

    foocor=1.0
    footype=instrument.getFootype()
    beamw=instrument.getBeamw()
    samlen=instrument.getSamplelen()
    if footype==0 or footype==instrument_string_choices['footype'][0]:
        foocor=1.0
    elif footype==1 or footype==instrument_string_choices['footype'][1]:
        foocor=GaussIntensity(theta, samlen/2.0, samlen/2.0, beamw)
    elif footype==2 or footype==instrument_string_choices['footype'][2]:
        foocor=SquareIntensity(theta, samlen, beamw)
    else:
        raise ValueError('Variable footype has an unvalid value')

    if restype==0 or restype==instrument_string_choices['restype'][0]:
        R=R[:]*foocor
    elif restype==1 or restype==instrument_string_choices['restype'][1]:
        R=ConvoluteFast(TwoThetaQz, R[:]*foocor, instrument.getRes(),
                        range=instrument.getResintrange())
    elif restype==2 or restype==instrument_string_choices['restype'][2]:
        R=ConvoluteResolutionVector(TwoThetaQz, R[:]*foocor, weight)
    elif restype==3 or restype==instrument_string_choices['restype'][3]:
        R=ConvoluteFastVar(TwoThetaQz, R[:]*foocor, instrument.getRes(),
                           range=instrument.getResintrange())
    else:
        raise ValueError('Variable restype has an unvalid value')
    return R+instrument.getIbkg()

def OffSpecularMingInterdiff(TwoThetaQz, ThetaQx, sample, instrument):
    ''' Function that simulates the off-specular signal (not implemented)
    
    # BEGIN Parameters
    TwoThetaQz 1.0
    ThetaQx data.x
    # END Parameters
    '''
    lamda=instrument.getWavelength()
    if instrument.getCoords() in [1, instrument_string_choices['coords'][1]]:
        alphaR1=ThetaQx
        betaR1=TwoThetaQz-ThetaQx
        qx=2*pi/lamda*(cos(alphaR1*pi/180)-cos(betaR1*pi/180))
        qz=2*pi/lamda*(sin(alphaR1*pi/180)+sin(betaR1*pi/180))
    else:
        qz=TwoThetaQz
        qx=ThetaQx

    parameters=sample.resolveLayerParameters()

    def toarray(a, code):
        a=list(a)
        a.reverse()
        return array(a, dtype=code)

    dens=array(parameters['dens'], dtype=complex64)
    f=array(parameters['f'], dtype=complex64)
    re=2.82e-13*1e2/1e-10
    n=1-dens*re*lamda**2/2/pi*f*1e-4
    n=toarray(n, code=complex64)
    sigmar=toarray(parameters['sigmar'], code=float64)
    sigmar=sigmar[1:]

    sigmai=toarray(parameters['sigmai'], code=float64)
    sigmai=sigmai[1:]+1e-5

    d=toarray(parameters['d'], code=float64)
    d=r_[0, d[1:-1]]

    z=-cumsum(d)

    eta=sample.getEta_x()

    h=sample.getH()

    eta_z=sample.getEta_z()

    (I, alpha, omega)=offspec.DWBA_Interdiff(qx, qz, lamda, n, z,
                                                 sigmar, sigmai, eta, h, eta_z, d,
                                                 taylor_n=instrument.getTaylor_n())

    restype=instrument.getRestype()
    if restype==0 or restype==instrument_string_choices['restype'][0]:
        # if no resolution is defined, don't include specular peak
        return real(I)*instrument.getI0()+instrument.getIbkg()

    # include specular peak
    instrument.setRestype(0)
    if isinstance(TwoThetaQz, ndarray):
        Ispec=Specular(TwoThetaQz, sample, instrument)
    else:
        Ispec=Specular(array([TwoThetaQz], dtype=float64), sample, instrument)[0]
    instrument.setRestype(restype)

    if instrument.getCoords() in [1, instrument_string_choices['coords'][1]]:
        spec_peak=Ispec*exp(-0.5*(TwoThetaQz/2.-ThetaQx)**2/instrument.getRes()**2)
    else:
        spec_peak=Ispec*exp(-0.5*ThetaQx**2/instrument.getRes()**2)
    return (spec_peak+real(I))*instrument.getI0()+instrument.getIbkg()

def SLD_calculations(z, item, sample, inst):
    ''' Calculates the scatteringlength density as at the positions z
    
    # BEGIN Parameters
    z data.x
    item "Re"
    # END Parameters
    '''
    parameters=sample.resolveLayerParameters()
    dens=array(parameters['dens'], dtype=complex64)
    f=array(parameters['f'], dtype=complex64)
    sld=dens*f
    d_sld=sld[:-1]-sld[1:]
    d=array(parameters['d'], dtype=float64)
    d=d[1:-1]
    # Include one extra element - the zero pos (substrate/film interface)
    int_pos=cumsum(r_[0, d])
    sigmar=array(parameters['sigmar'], dtype=float64)
    sigmar=sigmar[:-1]
    sigmai=array(parameters['sigmai'], dtype=float64)
    sigmai=sigmai[:-1]
    sigma=sqrt(sigmai**2+sigmar**2)+1e-7
    if z is None:
        z=arange(-sigma[0]*5, int_pos.max()+sigma[-1]*5, 0.5)
    rho=sum(d_sld*(0.5-0.5*erf((z[:, newaxis]-int_pos)/sqrt(2.)/sigma)), 1)+sld[-1]
    dic={'Re': real(rho), 'Im': imag(rho), 'z': z,
         'SLD unit': 'r_{e}/\AA^{3}'}
    if item is None or item=='all':
        return dic
    else:
        try:
            return dic[item]
        except:
            raise ValueError('The chosen item, %s, does not exist'%item)

SimulationFunctions={'Specular': Specular,
                     'OffSpecular': OffSpecularMingInterdiff,
                     'SLD': SLD_calculations}

from .lib import refl as Refl

(Instrument, Layer, Stack, Sample)=Refl.MakeClasses(InstrumentParameters,
                                                    LayerParameters, StackParameters,
                                                    SampleParameters, SimulationFunctions, ModelID)

if __name__=='__main__':
    Fe=Layer(d=10, sigmar=3.0, n=1-2.247e-5+2.891e-6j)
    Si=Layer(d=15, sigmar=3.0, n=1-7.577e-6+1.756e-7j)
    sub=Layer(sigmar=3.0, n=1-7.577e-6+1.756e-7j)
    amb=Layer(n=1.0, sigmar=1.0)
    stack=Stack(Layers=[Fe, Si], Repetitions=20)
    sample=Sample(Stacks=[stack], Ambient=amb, Substrate=sub, eta_z=500.0, eta_x=100.0)
    iprint(sample)
    inst=Instrument(Wavelength=1.54, Coordinates=1)



'''
===========================================
:mod:`orso` SINQ SIX standard file from PSI
===========================================

Load the simple text format used at the PSI SINQ neutron source.
Reads all clumns but uses s2t as standard x.
'''

import numpy as np
import re

from typing import List
from ..data_loader_framework import Template
from ..utils import ShowWarningDialog
from orsopy.fileio import load_orso, OrsoDataset

class Plugin(Template):
    wildcard='*.dat'

    def CanOpen(self, file_path):
        if not Template.CanOpen(self, file_path):
            return False
        l1=open(file_path, 'r', encoding='utf-8').readline()
        return 'Data File' in l1

    def LoadData(self, dataset, filename, data_id=0):
        '''
        Loads the data from filename into the data_item_number.
        '''
        try:
            lines=open(filename, 'r').readlines()
        except Exception as e:
            ShowWarningDialog(self.parent, 'Could not load the file: '+filename+
                              ' \nPlease check the format.\n\n ORSO reader gave the following error:\n'+str(e))
            return
        else:
            header=[]
            while len(lines)>0:
                ln=lines.pop(0)
                if '******* DATA ********' in ln:
                    lines.pop(0);lines.pop(0)
                    cols=[c.strip().lower() for c in lines.pop(0).split()]
                    break
                else:
                    header.append(ln.strip())
            while len(lines)>0:
                le=lines.pop(-1)
                if 'END-OF-DATA' in le:
                    break
            data=np.array([li.split() for li in lines], dtype=float).T


            dataset.x_raw=data[0]
            dataset.y_raw=data[cols.index('counts')]/data[cols.index('monitor1')]
            dataset.error_raw=np.sqrt(data[cols.index('counts')])/data[cols.index('monitor1')]

            for i, col in enumerate(data[1:]):
                colname=cols[i++1]
                if not colname.isidentifier() or colname in ['lambda', 'x', 'y', 'error', 'res']:
                    colname='col_'+re.sub('[^0-9a-zA-Z_]', '', colname)
                dataset.set_extra_data(colname, np.asarray(col), colname)
            # Name the dataset accordign to file name
            dataset.name='#'+filename[:-4].rsplit('n',1)[1]

            if 's2t' in cols:
                dataset.x_command='s2t'
            elif 'som' in cols:
                dataset.x_command='2.*som'

            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta['data_source']['measurement']['data_file_header']='\n'.join(header)



"""
GenX model and optimizer control classes. All functional aspects should be covered here (no GUI).
"""
import os
import sys
from typing import List

import h5py
import numpy as np
from logging import warning

from .core.config import config
from .data import DataList
from .exceptions import ErrorBarError, GenxIOError
from .model import Model
from .model_actions import ActionHistory, DeleteParams, ExchangeModel, InsertParam, ModelAction, MoveParam, NoOp, \
    UpdateModelScript, SortAndGroupParams, UpdateColorCycle, UpdateParams, UpdateParamValue, \
    UpdateSolverOptoins, UpdateDataPlotSettings
from .solver_basis import GenxOptimizer, GenxOptimizerCallback
from .plugins.data_loader_framework import Template as DataLoaderTemplate


class ModelController:
    model: Model
    optimizer: GenxOptimizer
    history: ActionHistory
    model_store: List[Model]

    def __init__(self, optimizer: GenxOptimizer):
        self.model = Model()
        self.optimizer = optimizer
        self.history = ActionHistory()
        self.model.saved = True
        self.model_store = []

    def action_callback(self, action: ModelAction):
        pass

    def set_callbacks(self, callbacks: GenxOptimizerCallback):
        self.optimizer.set_callbacks(callbacks)

    def set_action_callback(self, func):
        # sets a function to be called when aver an action is applied.
        self.action_callback = func

    def perform_action(self, action_class: type(ModelAction), *params):
        action = action_class(self.model, *params)
        self.history.execute(action)
        self.action_callback(action)

    def undo_action(self):
        action = self.history.undo()
        self.action_callback(action)
        return action

    def redo_action(self):
        action = self.history.redo()
        self.action_callback(action)
        return action

    def history_stacks(self):
        return self.history.undo_stack, self.history.redo_stack

    def history_clear(self):
        self.history.clear()
        self.action_callback(NoOp(None))

    def history_remove(self, start, length=1):
        actions = self.history.remove_actions(start, length)
        self.action_callback(actions)

    def is_configured(self):
        return self.optimizer.is_configured()

    def get_result_info(self):
        return self.optimizer.get_result_info()

    def get_fitted_model(self):
        return self.optimizer.get_model()

    def get_fom_log(self):
        return self.optimizer.get_fom_log()

    def ReadConfig(self):
        '''
        Reads the parameter that should be read from the config file.
        And set the parameters in both the optimizer and this class.
        '''
        self.model.ReadConfig()
        self.optimizer.ReadConfig()

    def WriteConfig(self):
        '''
        Writes the current configuration of the solver to file.
        '''
        self.model.WriteConfig()
        self.optimizer.WriteConfig()

    def new_model(self):
        self.model.new_model()
        self.history_clear()

    def set_model(self, model: Model):
        self.perform_action(ExchangeModel, model)

    def get_model(self) -> Model:
        return self.model

    def activate_model(self, store_index):
        '''
        Takes a model from the model_store list and sets it as active model_store.
        '''
        self.set_model(self.model_store[store_index])

    def put_in_store(self, store_index=None):
        '''
        Places a copy of the current model in
        '''
        if store_index is not None:
            self.model_store.insert(store_index, self.model.copy())
        else:
            self.model_store.append(self.model.deepcopy())

    def read_sequence(self, data_loader: DataLoaderTemplate, file_lists,
                      name_by_file=False, callback=None):
        data = self.get_data()
        if len(data)!=len(file_lists):
            raise ValueError("You have to provide a list of datafiles for each dataset in the model")
        for i, fi in enumerate(zip(*file_lists)):
            if callback:
                callback(i)
            for j, dj in enumerate(data):
                data_loader.LoadDataset(dj, fi[j])
            if name_by_file:
                fname = os.path.basename(fi[0])
                self.model.h5group_name = fname
            else:
                self.model.h5group_name = f'sequence_{len(self.model_store):05}'
            self.model.sequence_value = float(len(self.model_store))
            self.put_in_store()
        # activate the first dataset
        self.activate_model(0)

    def active_index(self):
        try:
            return self.model_store.index(self.model)
        except ValueError:
            return -1

    def set_model_script(self, text):
        old_script = self.model.get_script().replace('\r\n', '\n').replace('\r', '\n').strip()
        if text.strip().replace('\r\n', '\n').replace('\r', '\n')==old_script:
            # nothing to do, same script
            return
        self.perform_action(UpdateModelScript, text)

    def get_model_script(self):
        return self.model.get_script()

    def set_model_params(self, params):
        self.model.parameters = params

    def get_model_params(self):
        return self.model.parameters

    def set_data(self, data: DataList):
        self.model.data = data

    def get_data(self) -> DataList:
        return self.model.data

    def set_data_plotsettings(self, indices, sim_par, data_par):
        self.perform_action(UpdateDataPlotSettings, indices, sim_par, data_par)

    def update_color_cycle(self, source):
        if source!=self.model.data.color_source:
            self.perform_action(UpdateColorCycle, source)

    def get_color_cycle(self):
        return self.model.data.color_source

    def get_parameters(self):
        return self.model.parameters

    def get_sim_pars(self):
        return self.model.get_sim_pars()

    def get_parameter_data(self, row):
        return self.model.parameters.get_data()[row]

    def get_parameter_name(self, row):
        return self.model.parameters.get_names()[row]

    def get_possible_parameters(self):
        return self.model.get_possible_parameters()

    def set_error_pars(self, error_values):
        self.model.parameters.set_error_pars(error_values)

    def set_value_pars(self, new_values):
        pars = self.model.parameters.get_value_pars()
        if all([pi==pj for pi, pj in zip(new_values, pars)]):
            return
        self.perform_action(UpdateParams, new_values)

    def set_parameter_value(self, row, col, value):
        if value==self.model.parameters.get_value(row, col):
            return
        self.perform_action(UpdateParamValue, row, col, value)

    def move_parameter(self, row, step):
        self.perform_action(MoveParam, row, step)

    def insert_parameter(self, row):
        self.perform_action(InsertParam, row)

    def delete_parameter(self, rows):
        self.perform_action(DeleteParams, rows)

    def sort_and_group_parameters(self, sort_params):
        self.perform_action(SortAndGroupParams, sort_params)

    def get_fom(self):
        return self.model.fom

    def get_fom_name(self):
        return self.model.fom_func.__name__

    def set_filename(self, filename):
        self.model.filename = filename

    def get_filename(self):
        return self.model.filename

    def get_model_name(self):
        module = self.script_module
        name = module.model.__name__.rsplit('.', 1)[1]
        return name

    def force_compile(self):
        self.model.compile_script()

    def compile_if_needed(self):
        if not self.model.compiled:
            self.model.compile_script()

    def simulate(self, recompile=False):
        self.model.simulate(compile=(recompile or not self.model.compiled))

    def export_data(self, basename):
        self.model.export_data(basename)

    def export_table(self, basename):
        self.model.export_table(basename)

    def export_script(self, basename):
        self.model.export_script(basename)

    def export_orso(self, basename):
        self.model.export_orso(basename)

    def import_table(self, filename):
        self.model.import_table(filename)

    def import_script(self, filename):
        self.model.import_script(filename)

    def get_data_as_asciitable(self, indices=None):
        return self.model.get_data_as_asciitable(indices=indices)

    def get_combined_options(self):
        # Return a configuration object with all parameters relevant for fitting
        return self.model.solver_parameters | self.optimizer.opt

    def update_combined_options(self, new_values: dict):
        combined_options = self.get_combined_options()
        difference = {}
        for key, value in new_values.items():
            if getattr(combined_options, key, None)!=value:
                difference[key] = value
        self.perform_action(UpdateSolverOptoins, self.optimizer, difference)

    @property
    def saved(self):
        return self.model.saved

    @saved.setter
    def saved(self, value):
        self.model.saved = value

    @property
    def eval_in_model(self):
        return self.model.eval_in_model

    @property
    def script_module(self):
        self.compile_if_needed()
        return self.model.script_module

    def CalcErrorBars(self):
        '''
        Method that calculates the errorbars for the fit that has been
        done. Note that the fit has to been conducted before this is run.
        '''
        if self.optimizer.n_fom_evals==0:
            raise ErrorBarError('Can not find any stored evaluations of the model in the optimizer.\n'
                                'Run a fit before calculating the errorbars.')
        if self.optimizer.get_start_guess() is not None and not self.optimizer.is_running():
            n_elements = len(self.optimizer.get_start_guess())
            error_values = []
            for index in range(n_elements):
                # calculate the error, this is threshold based and not rigours
                error_values.append(self.optimizer.calc_error_bar(index))
            return np.array(error_values)
        else:
            raise ErrorBarError('Wait for fit to finish or fit to changed model first.')

    def ProjectEvals(self, parameter: int):
        '''
        Projects the parameter number parameter on one axis and returns
        the fom values.
        '''
        row = self.model.parameters.get_pos_from_row(parameter)
        if self.optimizer.get_start_guess() is not None and not self.optimizer.is_running():
            return self.optimizer.project_evals(row)
        else:
            raise ErrorBarError()

    def StartFit(self):
        '''
        Function to start running the fit
        '''
        if not self.model.compiled:
            self.simulate(recompile=True)
        # Make sure that the config of the solver is updated..
        self.optimizer.ReadConfig()
        # Reset all the errorbars
        self.model.parameters.clear_error_pars()
        self.optimizer.start_fit(self.model)

    def StopFit(self):
        '''
        Function to stop a running fit
        '''
        self.optimizer.stop_fit()

    def ResumeFit(self):
        '''
        Function to resume the fitting after it has been stopped
        '''
        self.compile_if_needed()
        self.optimizer.ReadConfig()
        self.optimizer.resume_fit(self.model)

    def IsFitted(self):
        '''
        Returns true if a fit has been started otherwise False
        '''
        return self.optimizer.is_fitted()

    def save(self):
        self.save_file(self.model.get_filename())

    def save_file(self, fname: str, update_callback=None):
        """
        Saves objects model, optimiser and config into file fname
        """
        if fname.lower().endswith('.gx'):
            self.save_gx(fname)
        elif fname.lower().endswith('.hgx'):
            self.save_hgx(fname, update_callback=update_callback)
        else:
            raise GenxIOError('Wrong file ending, should be .gx or .hgx')
        self.model.filename = os.path.abspath(fname)
        self.model.saved = True

    def load_file(self, fname: str, update_callback=None):
        """
        Loads parameters from fname into model, optimizer and config
        """
        if fname.lower().endswith('.gx'):
            self.load_gx(fname)
        elif fname.lower().endswith('.hgx'):
            self.load_hgx(fname, update_callback=update_callback)
        else:
            raise GenxIOError('Wrong file ending, should be .gx or .hgx')
        self.model.filename = os.path.abspath(fname)
        self.model.saved = True
        self.history_clear()

    def save_hgx(self, fname: str, update_callback=None):
        f = h5py.File(fname.encode('utf-8'), 'w')
        g = f.create_group('current')
        self.model.write_h5group(g)
        opt_group = g.create_group(self.optimizer.h5group_name)
        opt_group['solver'] = self.optimizer.__class__.__name__
        opt_group['solver_module'] = self.optimizer.__class__.__module__
        self.optimizer.write_h5group(opt_group)
        self.WriteConfig()
        g['config'] = config.model_dump().encode('utf-8')
        N = len(self.model_store)+1
        lN = len(f'{N}')
        fmt = '%%0%ii-%%s'%lN  # put an index before the name to keep list order on load
        for i, modeli in enumerate(self.model_store):
            if update_callback:
                update_callback(i+1, N)
            g = f.create_group(fmt%(i, modeli.h5group_name))
            modeli.write_h5group(g)
        f.close()

    def load_hgx(self, fname: str, update_callback=None):
        f = h5py.File(fname.encode('utf-8'), 'r')
        g = f['current']
        self.model.read_h5group(g)
        opt_group = g[self.optimizer.h5group_name]
        try:
            solver_class = opt_group.get('solver')[()]
            solver_module = opt_group.get('solver_module')[()]
        except (KeyError, TypeError):
            solver_class = 'DiffEv'
            solver_module = 'genx.diffev'
        else:
            if type(solver_class) is bytes:
                solver_class = solver_class.decode('utf-8')
                solver_module = solver_module.decode('utf-8')
        if solver_class!=self.optimizer.__class__.__name__:
            try:
                exec(f'from {solver_module} import {solver_class}')
            except ImportError:
                warning(f'Could not import solver {solver_class} from moudle {solver_module}')
            else:
                prev_optimizer = self.optimizer
                exec(f'self.optimizer={solver_class}()')
                self.optimizer.set_callbacks(prev_optimizer.get_callbacks())
        self.optimizer.read_h5group(opt_group)
        try:
            config.load_string(g['config'][()].decode('utf-8'))
            self.ReadConfig()
        except KeyError:
            pass
        except AttributeError:
            pass
        self.model_store = []
        N = len(f)
        names = [fi for fi in f.keys() if fi!='current']
        names.sort()  # make sure
        for i, gname in enumerate(names):
            if update_callback:
                update_callback(i, N)
            g = f[gname]
            modeli = Model()
            modeli.h5group_name = gname.split('-', 1)[-1]
            modeli.read_h5group(g)
            self.model_store.append(modeli)
        f.close()

    def save_gx(self, fname: str):
        self.model.save(fname)
        self.model.save_addition('config', config.model_dump())
        self.model.save_addition('optimizer', self.optimizer.pickle_string(clear_evals=
                                                                           not config.getboolean('solver',
                                                                                                 'save all evals')))

    def load_gx(self, fname: str):
        self._patch_modules()  # for compatibility with old files
        self.model.load(fname)
        config.load_string(self.model.load_addition('config').decode('utf-8'))
        self.optimizer.pickle_load(self.model.load_addition('optimizer'))

    def _patch_modules(self):
        # add legacy items to genx for loading of pickled strings from old program
        from genx import diffev
        diffev.defualt_autosave = diffev.DiffEv._callbacks.autosave



'''
Library for reflectivity calculations with neutrons.
'''
from numpy import *
from . import math_utils as mu
from . import int_lay_xmean
from functools import reduce
from genx.core.custom_logging import iprint


CTYPE=complex128

def ass_X_2int(k_mj, k_mj1, k_pj, k_pj1, theta_diff):
    costd=cos(theta_diff/2.0)
    sintd=sin(theta_diff/2.0)
    X=zeros((4, 4)+k_pj.shape, dtype=CTYPE)
    X[0, 0]=costd*(k_pj1+k_pj)/2./k_pj
    X[0, 1]=-costd*(k_pj1-k_pj)/2./k_pj
    X[0, 2]=sintd*(k_pj+k_mj1)/2./k_pj
    X[0, 3]=sintd*(k_pj-k_mj1)/2./k_pj
    # X[0] = X[0]/2/k_pj
    X[1, 0]=X[0, 1]  # -(costd*(k_pj1 - k_pj))/(2*k_pj)
    X[1, 1]=X[0, 0]  # (costd*(k_pj1 + k_pj))/(2*k_pj)
    X[1, 2]=X[0, 3]  # (sintd*(k_pj - k_mj1))/(2*k_pj)
    X[1, 3]=X[0, 2]  # (sintd*(k_pj + k_mj1))/(2*k_pj)
    X[2, 0]=-(sintd*(k_pj1+k_mj))/(2.*k_mj)
    X[2, 1]=(sintd*(k_pj1-k_mj))/(2.*k_mj)
    X[2, 2]=(costd*(k_mj1+k_mj))/(2.*k_mj)
    X[2, 3]=-(costd*(k_mj1-k_mj))/(2.*k_mj)
    X[3, 0]=X[2, 1]  # (sintd*(k_pj1 - k_mj))/(2*k_mj)
    X[3, 1]=X[2, 0]  # -(sintd*(k_pj1 + k_mj))/(2*k_mj)
    X[3, 2]=X[2, 3]  # -(costd*(k_mj1 - k_mj))/(2*k_mj)
    X[3, 3]=X[2, 2]  # (costd*(k_mj1 + k_mj))/(2*k_mj)
    return X

def ass_X(k_p, k_m, theta_diff):
    ''' Make the interface transmission matrix for neutron reflection from 
    a interface.
    '''
    # First is the substrate and last is the ambient (j=0)
    # The order is then [N .. j+1, j, ....0]
    k_pj1=k_p[:, :-1]
    k_pj=k_p[:, 1:]
    k_mj1=k_m[:, :-1]
    k_mj=k_m[:, 1:]
    return ass_X_2int(k_mj, k_mj1, k_pj, k_pj1, theta_diff)

def gauss(q, sigma2):
    '''Fourier transform of the interface roughness weight function
    '''
    return exp(-q**2*sigma2/2.0)

def include_sigma(X, k_p, k_m, sigma, w=gauss):
    '''Function to include roughness into the interface matrix.
    '''
    sigma2=sigma[..., :-1]**2
    k_pj1=k_p[:, :-1]
    k_pj=k_p[:, 1:]
    k_mj1=k_m[:, :-1]
    k_mj=k_m[:, 1:]
    X[0, 0]=X[0, 0]*w(k_pj-k_pj1, sigma2)
    X[0, 1]=X[0, 1]*w(k_pj+k_pj1, sigma2)
    X[0, 2]=X[0, 2]*w(k_pj-k_mj1, sigma2)
    X[0, 3]=X[0, 3]*w(k_pj+k_mj1, sigma2)
    X[1, 0]=X[0, 1]  # X[1,0]*w(k_pj + k_pj1, sigma2)
    X[1, 1]=X[0, 0]  # X[1,1]*w(k_pj - k_pj1, sigma2)
    X[1, 2]=X[0, 3]  # X[1,2]*w(k_pj + k_mj1, sigma2)
    X[1, 3]=X[0, 2]  # X[1,3]*w(k_pj - k_mj1, sigma2)
    X[2, 0]=X[2, 0]*w(k_mj-k_pj1, sigma2)
    X[2, 1]=X[2, 1]*w(k_mj+k_pj1, sigma2)
    X[2, 2]=X[2, 2]*w(k_mj-k_mj1, sigma2)
    X[2, 3]=X[2, 3]*w(k_mj+k_mj1, sigma2)
    X[3, 0]=X[2, 1]  # X[3,0]*w(k_mj + k_pj1, sigma)
    X[3, 1]=X[2, 0]  # X[3,1]*w(k_mj - k_pj1, sigma)
    X[3, 2]=X[2, 3]  # X[3,2]*w(k_mj + k_mj1, sigma)
    X[3, 3]=X[2, 2]  # X[3,3]*w(k_mj - k_mj1, sigma)
    return X

def ass_P(k_p, k_m, d):
    ''' Make the layer proagation matrix for a layer.
    '''
    P=zeros((4, 4)+k_p.shape, dtype=CTYPE)
    P[0, 0]=exp(-1.0J*k_p*d)
    P[1, 1]=1/P[0, 0]  # exp(1.0J*k_p*d)
    P[2, 2]=exp(-1.0J*k_m*d)
    P[3, 3]=1/P[2, 2]  # exp(1.0J*k_m*d)
    return P

def Refl(Q, Vp, Vm, d, M_ang, sigma=None, return_int=True):
    '''A quicker implementation than the ordinary slow implementaion in Refl
    Calculates spin-polarized reflectivity according to S.J. Blundell 
        and J.A.C. Bland Phys rev. B. vol 46 3391 (1992)
        The algorithm assumes that the first element in the arrays represents
        the substrate and the last the ambient layer.
        Input parameters:   Q : Scattering vector in reciprocal 
                                angstroms Q=4*pi/lambda *sin(theta)
                            Vp: Neutron potential for spin up
                            Vm: Neutron potential for spin down
                            d: layer thickness
                            M_ang: Angle of the magnetic 
                                    moment(radians!) M_ang=0 =>M//neutron spin
                            sigma: The roughness of the upper interface.
                            return_int: Flag for returning the instensity, default=True. If False return the amplitudes.
        Returns:            (Ruu,Rdd,Rud,Rdu)
                            (up-up,down-down,up-down,down-up)
    '''
    # Assume first element=substrate and last=ambient!
    k_amb=Q[:, newaxis]/2.0
    Vp=Vp.astype(complex128);Vm=Vm.astype(complex128)
    if M_ang[-1]!=0:
        raise ValueError("The magnetization in the ambient layer has to be in polarization direction")
    if Vp[-1]!=0 or Vm[-1]!=0:
        # Ambient not vacuum
        raise ValueError("The SLD in the ambient layer has to be zero, apply renormalization first")
    if len(Vp)==2:
        # Algorithm breaks without a layer, so add an empty one
        Vp=hstack([Vp, [Vp[-1]]])
        Vm=hstack([Vm, [Vm[-1]]])
        M_ang=array([M_ang[0], 0., 0.], dtype=float64)
        d=array([d[0], 10., d[1]], dtype=float64)
        if sigma is not None:
            sigma=array([sigma[0], sigma[0], sigma[1]], dtype=float64)
    # Wavevectors in the layers
    k_p=sqrt(k_amb**2-Vp)
    k_m=sqrt(k_amb**2-Vm)
    # Angular difference between the magnetization
    theta_diff=M_ang[1:]-M_ang[:-1]
    # if sigma is None:
    #    sigma = zeros(d.shape)
    # Assemble the interface reflectivity matrix
    X=ass_X(k_p, k_m, theta_diff)
    if sigma is not None:
        X=include_sigma(X, k_p, k_m, sigma)
    # Assemble the layer propagation matrices
    P=ass_P(k_p, k_m, d)
    # Multiply the propagation matrices with the interface matrix
    PX=mu.dot4_Adiag(P[..., 1:-1], X[..., :-1])
    # Multiply up the sample matrix
    # print 'X: ', X[:,:, 0, -1]
    M=mu.dot4(X[..., -1], reduce(mu.dot4, rollaxis(PX, 3)[::-1]))
    # print M.shape
    # print 'M: ', M[:,:, 0]
    # print 'denom: ',  M[0,0]*M[2,2]-M[0,2]*M[2,0]
    denom=M[0, 0]*M[2, 2]-M[0, 2]*M[2, 0]
    Ruu=(M[1, 0]*M[2, 2]-M[1, 2]*M[2, 0])/denom
    Rud=(M[3, 0]*M[2, 2]-M[3, 2]*M[2, 0])/denom
    Rdu=(M[1, 2]*M[0, 0]-M[1, 0]*M[0, 2])/denom
    Rdd=(M[3, 2]*M[0, 0]-M[3, 0]*M[0, 2])/denom

    if return_int:
        return abs(Ruu)**2, abs(Rdd)**2, abs(Rud)**2, abs(Rdu)**2
    else:
        return Ruu, Rdd, Rud, Rdu

def Refl_int_lay(Q, V0, Vmag, d, M_ang, sigma, dmag_u, dd_u, M_ang_u, sigma_u,
                 dmag_l, dd_l, M_ang_l, sigma_l, return_int=True):
    '''A quicker implementation than the ordinary slow implementaion in Refl
    Calculates spin-polarized reflectivity according to S.J. Blundell
        and J.A.C. Bland Phys rev. B. vol 46 3391 (1992)
        The algorithm assumes that the last element in the arrays represents
        the substrate and the first the ambient layer.
        Input parameters:   Q : Scattering vector in reciprocal
                                angstroms Q=4*pi/lambda *sin(theta)
                            Vp: Neutron potential for spin up
                            Vm: Neutron potential for spin down
                            d: layer thickness
                            M_ang: Angle of the magnetic
                                    moment(radians!) M_ang=0 =>M//nuetron spin
                            sigma: The roughness of the upper interface.
                            The subscript l and u denotes the lower and upper interface
                            , respectively.
        Returns:            (Ruu,Rdd,Rud,Rdu)
                            (up-up,down-down,up-down,down-up)
    '''
    Vp=V0+Vmag
    Vm=V0-Vmag
    Vp_u=V0+Vmag*(1.+dmag_u)
    Vm_u=V0-Vmag*(1.+dmag_u)
    Vp_l=V0+Vmag*(1.+dmag_l)
    Vm_l=V0-Vmag*(1.+dmag_l)
    # Assume last element=substrate and first=ambient!
    k_amb=Q[:, newaxis]/2.0
    # Wavevectors in the layers
    kp=sqrt(k_amb**2-Vp).astype(complex128)
    km=sqrt(k_amb**2-Vm).astype(complex128)
    kp_u=sqrt(k_amb**2-Vp_u).astype(complex128)
    km_u=sqrt(k_amb**2-Vm_u).astype(complex128)
    kp_l=sqrt(k_amb**2-Vp_l).astype(complex128)
    km_l=sqrt(k_amb**2-Vm_l).astype(complex128)
    # Angular difference between the magnetization
    theta_diff=M_ang[:-1]-M_ang[1:]
    theta_diff_lu=M_ang_l[:-1]-M_ang_u[1:]
    theta_diff_l=M_ang[:-1]-M_ang_l[:-1]
    theta_diff_u=M_ang_u[1:]-M_ang[1:]

    X_lu=ass_X_2int(km_l[:, :-1], km_u[:, 1:], kp_l[:, :-1], kp_u[:, 1:], theta_diff_lu)
    X_l=ass_X_2int(km[:, :-1], km_l[:, :-1], kp[:, :-1], kp_l[:, :-1], theta_diff_l)
    X_u=ass_X_2int(km_u[:, 1:], km[:, 1:], kp_u[:, 1:], kp[:, 1:], theta_diff_u)

    X=int_lay_xmean.calc_neu_Xmean(X_l, X_lu, X_u, km, kp, km_l, kp_l, km_u, kp_u, dd_u, dd_l, sigma, sigma_l, sigma_u)

    # Assemble the layer propagation matrices
    P=ass_P(kp, km, d-dd_u-dd_l)
    # Multiply the propagation matrices with the interface matrix
    PX=mu.dot4_Adiag(P[..., 1:-1], X[..., 1:])
    # Multiply up the sample matrix
    M=mu.dot4(X[..., 0], reduce(mu.dot4, rollaxis(PX, 3)))
    # print M.shape
    denom=M[0, 0]*M[2, 2]-M[0, 2]*M[2, 0]
    Ruu=(M[1, 0]*M[2, 2]-M[1, 2]*M[2, 0])/denom
    Rud=(M[3, 0]*M[2, 2]-M[3, 2]*M[2, 0])/denom
    Rdu=(M[1, 2]*M[0, 0]-M[1, 0]*M[0, 2])/denom
    Rdd=(M[3, 2]*M[0, 0]-M[3, 0]*M[0, 2])/denom

    if return_int:
        return abs(Ruu)**2, abs(Rdd)**2, abs(Rud)**2, abs(Rdu)**2
    else:
        return Ruu, Rdd, Rud, Rdu

from . import USE_NUMBA

if USE_NUMBA:
    # try to use numba to speed up the calculation intensive functions:
    try:
        from .neutron_numba import Refl
    except Exception as e:
        iprint('Could not use numba, no speed up from JIT compiler:\n'+str(e))



'''<h1> SXRD plugin </h1>
This plugin auto generates the sample definition and simulations of a surface x-ray diffraction experiment.

'''

import os
import io
import traceback
from logging import debug

import wx

from .. import add_on_framework as framework
from .help_modules import model_interactors as mi
from .help_modules import atom_viewer
from genx.core.custom_logging import iprint
from genx.core.decorators import log_call
from genx.gui.solvergui import EVT_UPDATE_SCRIPT

code="""
        # BEGIN Instruments
        inst = model.Instrument(wavel=0.64, alpha=0.5, geom="alpha_in fixed")
        # END Instruments

        # BEGIN UnitCells
        uc = model.UnitCell(a=6, b=6, c=6, alpha=90, beta=90, gamma=90)
        # END UnitCells

        # BEGIN Slabs
        slab = model.Slab(c=1.0)
        slab.add_atom(id='al', el='Al', x=0.5, y=0.5, z=0.5, u=1.0, m=1, oc=1.0)
        slab.add_atom(id='la', el='La', x=0.0, y=0.0, z=0.0, u=0.05, m=4, oc=0.5)

        slab2 = model.Slab(c=1.0)
        slab2.add_atom(id='Y', el='Y', x=0.5, y=0.5, z=0.5, u=1.0, m=1, oc=1.0)
        slab2.add_atom(id='Cu1+', el='Cu', x=0.0, y=0.0, z=0.0, u=0.05, m=4, oc=0.5)
        # END Slabs

        # BEGIN Domains
        domain1 = model.Domain(slab, [slab, slab2], uc, occ=1.0, surface_sym=p1, bulk_sym=p1)
        domain2 = model.Domain(slab, [slab, slab, slab, slab2, slab2], uc, occ=1.0, surface_sym=p1, bulk_sym=p1)
        # END Domains

        # BEGIN Samples
        sample = model.Sample([domain1, domain2], cohf=0.0)
        # END Samples

        # BEGIN Parameters
        cp = UserVars()
        cp.new_var('test', 0.0)
        # END Parameters

        def Sim(data):
            I = []
            # BEGIN DataSet 0
            d = data[0]
            I.append(sample.calc_i(inst, 0, 0, d.x))
            # END DataSet 0
            return I
        """

class Plugin(framework.Template):
    def __init__(self, parent):

        framework.Template.__init__(self, parent)
        self.setup_script_interactor()
        if self.GetModelScript()=='':
            self.script_interactor.parse_code(code)
            self.SetModelScript(self.script_interactor.get_code())
        else:
            try:
                self.script_interactor.parse_code(self.GetModelScript())
            except Exception as e:
                iprint("SXRD plugin model could not be read.")
                self.script_interactor.parse_code(code)
                self.SetModelScript(self.script_interactor.get_code())

        self.layout_sample_edit()
        self.layout_simulation_edit()
        self.layout_misc_edit()
        self.layout_domain_viewer()
        self.create_main_window_menu()

        self.OnInteractorChanged(None)
        self.update_data_names()
        self.simulation_edit_widget.Update()
        self.update_widgets()

        parent.model_control.Bind(EVT_UPDATE_SCRIPT, self.ReadUpdateModel)


    @log_call
    def ReadUpdateModel(self, evt):
        self.script_interactor.parse_code(self.GetModelScript())
        self.OnInteractorChanged(None)
        self.update_data_names()
        self.simulation_edit_widget.Update()
        self.update_widgets()

    def setup_script_interactor(self, model_name='sxrd2'):
        """Setup the script interactor"""
        model=__import__('models.%s'%model_name, globals(), locals(), [model_name])
        preamble="import models.%s as model\nfrom models.utils import UserVars\nfrom models.symmetries import *\n"%model_name
        script_interactor=mi.ModelScriptInteractor(preamble=preamble)

        script_interactor.add_section('Instruments', mi.ObjectScriptInteractor, class_name='model.Instrument',
                                      class_impl=model.Instrument)
        script_interactor.add_section('UnitCells', mi.ObjectScriptInteractor, class_name='model.UnitCell',
                                      class_impl=model.UnitCell)
        script_interactor.add_section('Slabs', mi.SlabInteractor, class_name='model.Slab', class_impl=model.Slab)
        script_interactor.add_section('Domains', mi.DomainInteractor, class_name='model.Domain',
                                      class_impl=model.Domain)
        script_interactor.add_section('Samples', mi.SampleInteractor, class_name='model.Sample',
                                      class_impl=model.Sample)

        self.script_interactor=script_interactor

    @log_call
    def layout_sample_view(self):
        """Layouts the sample_view_panel"""
        panel=self.NewPlotFolder('Test')
        sizer=wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        self.sample_edit_widget=mi.DomainListCtrl(panel, domain_list=self.script_interactor.domains,
                                                  slab_list=self.script_interactor.slabs,
                                                  unitcell_list=self.script_interactor.unitcells)
        panel.Bind(mi.EVT_INTERACTOR_CHANGED, self.OnInteractorChanged, self.sample_edit_widget)
        sizer.Add(self.sample_edit_widget, 1, wx.EXPAND)
        panel.Layout()

    @log_call
    def layout_sample_edit(self):
        """Layouts the sample_edit_panel"""
        panel=self.NewInputFolder('Sample')
        sizer=wx.BoxSizer(wx.HORIZONTAL)
        panel.SetSizer(sizer)
        self.sample_edit_widget=mi.DomainListCtrl(panel, domain_list=self.script_interactor.domains,
                                                  slab_list=self.script_interactor.slabs,
                                                  unitcell_list=self.script_interactor.unitcells,
                                                  sample_list=self.script_interactor.samples)
        panel.Bind(mi.EVT_INTERACTOR_CHANGED, self.OnInteractorChanged, self.sample_edit_widget)
        panel.Bind(mi.EVT_SELECTION_CHANGED, self.OnSelectionChanged, self.sample_edit_widget)

        sizer.Add(self.sample_edit_widget, 1, wx.EXPAND)
        panel.Layout()

    @log_call
    def layout_simulation_edit(self):
        """Layouts the simulation_edit_panel"""
        panel=self.NewInputFolder('Simulations')
        sizer=wx.BoxSizer(wx.HORIZONTAL)
        panel.SetSizer(sizer)
        self.simulation_edit_widget=mi.SimulationListCtrl(panel, self, self.script_interactor)
        panel.Bind(mi.EVT_INTERACTOR_CHANGED, self.OnInteractorChanged, self.simulation_edit_widget)
        sizer.Add(self.simulation_edit_widget, 1, wx.EXPAND)
        panel.Layout()

    @log_call
    def layout_misc_edit(self):
        """Layouts the misc_edit_panel"""
        panel=self.NewDataFolder('Misc')
        sizer=wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        # Make the box for putting in the columns
        col_box=wx.StaticBox(panel, -1, "Unit Cells")
        col_box_sizer=wx.StaticBoxSizer(col_box, wx.VERTICAL)
        sizer.Add(col_box_sizer, 1, wx.EXPAND)
        self.unitcell_edit_widget=mi.EditList(panel, object_list=self.script_interactor.unitcells,
                                              default_name='Unitcells',
                                              edit_dialog=mi.ObjectDialog, edit_dialog_name='Unitcell Editor')
        panel.Bind(mi.EVT_INTERACTOR_CHANGED, self.OnInteractorChanged, self.unitcell_edit_widget)
        col_box_sizer.Add(self.unitcell_edit_widget, 1, wx.EXPAND)
        # Make the box for putting in the columns
        col_box=wx.StaticBox(panel, -1, "Instruments")
        col_box_sizer=wx.StaticBoxSizer(col_box, wx.VERTICAL)
        sizer.Add(col_box_sizer, 1, wx.EXPAND)
        self.instrument_edit_widget=mi.EditList(panel, object_list=self.script_interactor.instruments,
                                                default_name='Instruments',
                                                edit_dialog=mi.ObjectDialog, edit_dialog_name='Instrument Editor')
        panel.Bind(mi.EVT_INTERACTOR_CHANGED, self.OnInteractorChanged, self.instrument_edit_widget)
        col_box_sizer.Add(self.instrument_edit_widget, 1, wx.EXPAND)
        panel.Layout()

    @log_call
    def layout_domain_viewer(self):
        """Creates a 3D view of the sample."""
        panel=self.NewPlotFolder('Sample view')
        sample_view_sizer=wx.BoxSizer(wx.HORIZONTAL)
        panel.SetSizer(sample_view_sizer)
        self.sample_view=atom_viewer.VTKview(panel)
        self.sample_view.ReadConfig=lambda: None
        self.sample_view.GetZoom=lambda: False
        toolbar=self.sample_view.do_toolbar(panel)

        sample_view_sizer.Add(toolbar, 0, wx.EXPAND)
        sample_view_sizer.Add(self.sample_view, 1, wx.EXPAND | wx.GROW | wx.ALL)

        toolbar.Realize()

        # Just to init the view properly
        cur_page=self.parent.plot_notebook.Selection
        self.parent.plot_notebook.SetSelection(self.parent.plot_notebook.GetPageCount()-1)
        self.parent.plot_notebook.SetSelection(cur_page)
        self.sample_view.show()

    @log_call
    def create_main_window_menu(self):
        """Creates the window menu"""
        self.menu=self.NewMenu('SXRD')
        menu_item=wx.MenuItem(self.menu, wx.NewId(), "Domain Viewer/Export Settings...", "Edit Viewer settings",
                              wx.ITEM_NORMAL)
        self.menu.Append(menu_item)
        mb_export_xyz=wx.MenuItem(self.menu, wx.NewId(),
                                       "Export XYZ...",
                                       "Export the SLD to a XYZ ASCII file",
                                       wx.ITEM_NORMAL)
        self.menu.Append(mb_export_xyz)

        self.parent.Bind(wx.EVT_MENU, self.OnExportXYZ, mb_export_xyz)
        self.parent.Bind(wx.EVT_MENU, self.OnDomainViewerSettings, menu_item)

    def update_script(self):
        """Updates the script with new data"""
        old_script=self.GetModelScript()
        if old_script=="":
            self.SetModelScript(self.script_interactor.get_code())
        else:
            try:
                self.SetModelScript(self.script_interactor.update_code(old_script))
            except Exception as e:
                outp=io.StringIO()
                traceback.print_exc(200, outp)
                tbtext=outp.getvalue()
                outp.close()
                iprint("Error updating the script: ")
                iprint(tbtext)
                if self.ShowQuestionDialog('Could not update the script due to syntax issues. Python error: %s\n\n'
                                           'Do you wish to reset the model to the one defined in the user interface?'):
                    self.SetModelScript(self.script_interactor.get_code())

    def update_data_names(self):
        """Updates the DataSetInteractors names from the DataSet names in the model"""
        data_set_list=self.GetModel().data

        #assert (len(data_set_list)==len(self.script_interactor.data_sections_interactors))

        for interactor, data_set in zip(self.script_interactor.data_sections_interactors, data_set_list):
            interactor.set_name(data_set.name)

    def update_widgets(self):
        """Collects the already defined names and sets the already taken names in the different controls.
        Updates the symmetries from the script.
        """
        self.CompileScript()
        names=dir(self.GetScriptModule())
        self.instrument_edit_widget.set_taken_names(names[:])
        self.unitcell_edit_widget.set_taken_names(names[:])
        self.sample_edit_widget.set_taken_names(names[:])

        model=self.GetModel()
        symmetries=[obj for obj in names if model.eval_in_model('isinstance(%s, Sym)'%obj)]
        self.sample_edit_widget.set_symmetries(symmetries)

    def set_constant_names(self):
        """Sets the name that needs to constant (used in other defs)"""
        self.unitcell_edit_widget.set_undeletable_names([d.unitcell for d in self.script_interactor.domains])
        self.instrument_edit_widget.set_undeletable_names([ds.instrument for ds in
                                                           self.script_interactor.data_sections_interactors])

    def update_domain_view(self):
        domain=self.sample_edit_widget.get_selected_domain_name()
        if domain:
            try:
                domain=self.GetModel().eval_in_model(domain)
            except Exception:
                iprint("Could not load domain ", domain)
            else:
                self.sample_view.build_sample(domain, use_opacity=False)

    def OnDomainViewerSettings(self, event):
        """Callback for showing the Domain Viewer settings dialog"""
        self.sample_view.ShowSettingDialog()
        self.update_domain_view()

    def OnInteractorChanged(self, event):
        """Callback when an Interactor has been changed by the GUI"""
        self.update_script()
        self.set_constant_names()
        self.update_widgets()
        self.update_domain_view()

    def OnSelectionChanged(self, evnet):
        """Callback when the selection in the sample widget has changed"""
        self.update_domain_view()

    def OnNewModel(self, event):
        """Callback for creating a new model"""
        self.update_script()
        self.update_data_names()
        self.simulation_edit_widget.Update()

    def OnOpenModel(self, event):
        """Callback for opening a model"""
        self.script_interactor.parse_code(self.GetModelScript())
        self.update_data_names()
        self.simulation_edit_widget.Update()
        self.set_constant_names()

    def OnDataChanged(self, event):
        """Callback for changing of the data sets (dataset added or removed)"""
        # We have the following possible events:
        # event.new_model, event.data_moved, event.deleted, event.new_data, event.name_change
        if event.new_model:
            # If a new model is created bail out
            return

        if event.new_data and len(self.script_interactor.data_sections_interactors)<len(self.GetModel().data):
            # New data has been added:
            self.script_interactor.append_dataset()
        elif event.deleted:
            for pos in event.position:
                self.script_interactor.data_sections_interactors.pop(pos)
        elif event.data_moved:
            if event.up:
                for pos in event.position:
                    tmp=self.script_interactor.data_sections_interactors.pop(pos)
                    self.script_interactor.data_sections_interactors.insert(pos-1, tmp)
            else:
                for pos in event.position:
                    tmp=self.script_interactor.data_sections_interactors.pop(pos)
                    self.script_interactor.data_sections_interactors.insert(pos+1, tmp)

        self.update_data_names()
        self.simulation_edit_widget.Update()
        self.update_script()

    def OnSimulate(self, event):
        """Callback called after simulation"""
        pass

    def OnExportXYZ(self, event):
        domain=self.sample_edit_widget.get_selected_domain_name()
        if domain:
            try:
                domain_obj=self.GetModel().eval_in_model(domain)
            except Exception:
                iprint("Could not load domain ", domain)
                return
        else:
            iprint("No domain selected.")
            return

        dlg=wx.FileDialog(self.parent, message="Export Domain to XYZ file ...",
                          defaultFile=f"{self.GetModel().filename.rsplit('.',1)[0]}_{domain}.xyz",
                          wildcard="XYZ File (*.xyz)|*.xyz",
                          style=wx.FD_SAVE | wx.FD_CHANGE_DIR
                          )
        if dlg.ShowModal()==wx.ID_OK:
            fname=dlg.GetPath()
            result=True
            if os.path.exists(fname):
                filepath, filename=os.path.split(fname)
                result=self.ShowQuestionDialog('The file %s already exists.'
                                               ' Do'
                                               ' you wish to overwrite it?'
                                               %filename)
            if result:
                sv=self.sample_view
                domain_obj.export_xyz(fname, use_sym=sv.use_sym,
                                      x_uc=sv.x_uc, y_uc=sv.y_uc,
                                      fold_sym=sv.fold_sym, use_bulk=sv.show_bulk)


# Functions for instrumental resolution corrections and area of illumination
# these are to be included in the Models class and Refl class
# Note that these function assumes a Gaussian spread in both the angular as well
# as in real space

# THis one really needs scipy
from numpy import *
from scipy.special import erf
from genx.core.custom_logging import iprint


rad = pi/180.
sqrt2 = sqrt(2.)


# Intensity correction Gaussian beamprofile
def GaussIntensity(alpha, s1, s2, sigma_x):
    sinalpha = sin(alpha*rad)
    if s1==s2:
        return erf(s2/sqrt2/sigma_x*sinalpha)
    else:
        common = sinalpha/sqrt2/sigma_x
        return (erf(s2*common)+erf(s1*common))/2.0


# Specular foorprintcorrections square beamprofile
def SquareIntensity(alpha, slen, beamwidth):
    F = slen/beamwidth*sin(alpha*rad)
    return where(F<=1.0, F, ones(F.shape))


####################################################################
## Resolution Functions (Normal Distributed=Gaussian)
#####################################################################

# Full Convlutions - vayring resolution

# Function to create a 1D vector for the resolution with the 
# positions to calculate the reflectivity Qret and the weight
# of each point weight
# Inputs: Q - the Q values 
#         dQ - the resolution
#         points - the number of points for the convolution
#         range how far the gaussian should be convoluted
def ResolutionVector(Q, dQ, points, range=3):
    # if type(dQ)!=type(array([])):
    #    dQ=dQ*ones(Q.shape)
    Qres = Q+dQ*linspace(-range, range, points)[:, newaxis]

    weight = 1/sqrt(2*pi)/dQ*exp(-(transpose(Q[:, newaxis])-Qres)**2/dQ**2/2)
    Qret = Qres.flatten()  # reshape(Qres,(1,Qres.shape[0]*Qres.shape[1]))[0]
    # print Qres
    # print Qres.shape
    # print Qret.shape
    return Qret, weight


# Include the resolution with Qret and weight calculated from ResolutionVector
# and I the calculated intensity at each point. returns the intensity
def ConvoluteResolutionVector(Qret, I, weight):
    Qret2 = Qret.reshape(weight.shape[0], weight.shape[1])
    # print Qret.shape,weight.shape
    I2 = I.reshape(weight.shape[0], weight.shape[1])
    # print (I*weight).shape,Qret.shape
    norm_fact = trapz(weight, x=Qret2, axis=0)
    Int = trapz(I2*weight, x=Qret2, axis=0)/norm_fact
    # print Int.shape
    return Int


# Fast convlution - constant resolution
# constant spacing between data!
def ConvoluteFast(Q, I, dQ, range=3):
    Qstep = Q[1]-Q[0]
    resvector = arange(-range*dQ, range*dQ+Qstep, Qstep)
    weight = 1/sqrt(2*pi)/dQ*exp(-resvector**2/dQ**2/2)
    Iconv = convolve(r_[ones(resvector.shape)*I[0], I, ones(resvector.shape)*I[-1]], weight/weight.sum(),
                     mode='same')[resvector.shape[0]:-resvector.shape[0]]
    return Iconv


# Fast convolution - varying resolution
# constant spacing between the dat.
def ConvoluteFastVar(Q, I, dQ, range=3):
    Qstep = Q[1]-Q[0]
    steps = max(dQ*ones(Q.shape))*range/Qstep
    weight = 1/sqrt(2*pi)/dQ*exp(-(Q[:, newaxis]-Q)**2/dQ**2/2)
    Itemp = I[:, newaxis]*ones(I.shape)
    norm_fact = trapz(weight, axis=0)
    Int = trapz(Itemp*weight, axis=0)/norm_fact
    return Int


def QtoTheta(wavelength, Q):
    return arcsin(wavelength/4.0/pi*Q)/rad


def TwoThetatoQ(wavelength, TwoTheta):
    return 4.0*pi/wavelength*sin(rad/2.*TwoTheta)


########## Neutron Polarization #################
def get_pol_matrix(p1, p2, F1, F2):
    # @formatter:off
    P1 = array([
                [       1,       0,       0,      0],
                [       0,       1,       0,      0],
                [      F1,       0, (1.-F1),      0],
                [       0,      F1,       0, (1.-F1)],
               ])
    P2 = array([
                [       1,       0,       0,      0],
                [      F2, (1.-F2),       0,      0],
                [       0,       0,       1,      0],
                [       0,       0,      F2, (1.-F2)],
               ])
    P3 = array([
                [ (1.-p1),       0,      p1,       0],
                [       0, (1.-p1),       0,      p1],
                [      p1,       0, (1.-p1),       0],
                [       0,      p1,       0, (1.-p1)],
               ])
    P4 = array([
                [ (1.-p2),      p2,       0,       0],
                [      p2, (1.-p2),       0,       0],
                [       0,       0, (1.-p2),      p2],
                [       0,       0,      p2, (1.-p2)],
               ])
    # @formatter:on
    P = dot(dot(P1, P2), dot(P3, P4))
    return P


from . import USE_NUMBA


if USE_NUMBA:
    # try to use numba to speed up the calculation intensive functions:
    try:
        from .instrument_numba import GaussIntensity, SquareIntensity, QtoTheta, TwoThetatoQ, ResolutionVector
    except Exception as e:
        iprint('Could not use numba, no speed up from JIT compiler:\n'+str(e))



"""
Classes for handling of messages with console.

If available (UNIX) uses the curses library for a pseudo-interactive interface.
"""

import logging
import atexit
from numpy import array, arange

from . import config as io
from .custom_logging import iprint
from ..model_control import ModelController
from ..solver_basis import GenxOptimizer, GenxOptimizerCallback


class CursesHandler(logging.Handler):
    message_history = []
    max_messages = 8

    def __init__(self, optimizer: GenxOptimizer, *args, **kwargs):
        logging.Handler.__init__(self, *args, **kwargs)
        self.opt = optimizer
        import curses

        self.stdscr = curses.initscr()
        curses.noecho()
        curses.cbreak()
        self.stdscr.clear()

        _, width = self.stdscr.getmaxyx()
        self.stdscr.addstr(9, width//2-22, 'Relative value and spread of fit parameters:')
        self.stdscr.addstr(9, width-11, 'best/width')
        self.stdscr.nodelay(True)
        self.stdscr.refresh()

        rl = logging.getLogger()
        rl.handlers[0].setLevel(logging.ERROR+1)

        self.setLevel(logging.INFO)
        rl.addHandler(self)

        atexit.register(self.reset_console)

    def emit(self, record):
        try:
            msg = self.format(record)
            self.message_history += msg.strip().splitlines()
            lkey = self.stdscr.getch()
            if lkey==ord('q'):
                self.message_history.append('"q" pressed, trying to stop fit.')
                self.opt.stop = True
            self.message_history = self.message_history[-self.max_messages:]
            for i, msg in enumerate(self.message_history):
                _, width = self.stdscr.getmaxyx()
                self.stdscr.addstr(i, 4, msg+' '*(width-4-len(msg)))
            self.stdscr.refresh()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def reset_console(self):
        import curses
        self.setLevel(logging.ERROR+1)
        curses.nocbreak()
        curses.echo()
        curses.endwin()
        rl = logging.getLogger()
        rl.handlers[0].setLevel(logging.INFO)
        print('\n'.join(self.message_history))


def calc_errorbars(mod, opt: GenxOptimizer):
    error_values = []
    n_elements = len(opt.start_guess)
    for index in range(n_elements):
        # calculate the error for given optimizer, accuracy depends on settings
        (error_low, error_high) = opt.calc_error_bar(index)
        error_values.append('(%.3e, %.3e,)'%(error_low, error_high))
    mod.parameters.set_error_pars(error_values)


class ConsoleCallback(GenxOptimizerCallback):

    def __init__(self, stdscr, ctrl: ModelController, error=False, outfile=None):
        self.stdscr = stdscr
        self.ctrl = ctrl
        self.error = error
        self.outfile = outfile

    def autosave(self):
        # print 'Updating the parameters'
        model = self.ctrl.get_model()
        model.parameters.set_value_pars(self.ctrl.optimizer.best_vec)
        if self.error:
            iprint("Calculating error bars")
            calc_errorbars(self.ctrl.model, self.ctrl.optimizer)
        if self.outfile:
            iprint("Saving to %s"%self.outfile)
            self.ctrl.save_file(self.outfile)

    def parameter_output(self, param_info):
        if self.stdscr:
            height, width = self.stdscr.getmaxyx()
            full_width = (width-16-12)
            pwidth = param_info.max_val-param_info.min_val
            pres = (param_info.values-param_info.min_val)/pwidth
            population = array(param_info.population)
            pmin = (population.min(axis=0)-param_info.min_val)/pwidth
            pmax = (population.max(axis=0)-param_info.min_val)/pwidth
            pwidth = pmax-pmin
            maxpar = height-10
            if maxpar<param_info.values.shape[0]:
                order = pwidth.argsort()[::-1]
            else:
                order = arange(param_info.values.shape[0])

            for i in range(min(param_info.values.shape[0], maxpar)):
                param_id = order[i]
                self.stdscr.hline(10+i, 16, ' ', full_width)
                self.stdscr.addstr(10+i, 1, 'Parameter %02i: ['%param_id)
                self.stdscr.addstr(10+i, width-12, '] %.2f/%.2f'%(pres[param_id], pwidth[param_id]))

                ppos = max(0.0, min(1.0, pres[param_id]))
                wstart = max(0.0, min(1.0, pmin[param_id]))
                wwidth = max(0.0, min(1.0-wstart, pwidth[param_id]))
                self.stdscr.hline(10+i, 16+int(full_width*wstart), '=', int(full_width*wwidth))
                self.stdscr.addstr(10+i, 16+int(full_width*ppos), '#')

            self.stdscr.refresh()

    def text_output(self, text):
        iprint(text, flush=True)

    def plot_output(self, update_data):
        pass

    def fitting_ended(self, result_data):
        pass


def setup_console(ctrl: ModelController, error=False, outfile=None, use_curses=True):
    if use_curses:
        try:
            import curses
        except ImportError:
            stdscr = None
        else:
            chandler = CursesHandler(ctrl.optimizer)
            stdscr = chandler.stdscr
    else:
        stdscr = None

    cb = ConsoleCallback(stdscr, ctrl, error, outfile)
    ctrl.set_callbacks(cb)



'''<h1>Shell</h1>
A plugin which opens a up a new tab with a shell inside.
Can be used for introspection in GenX. To acess the module
which holds the script type: <code>model.script_module.[object]</code>.
This is mostly intended for searching for bugs and test expresions
and such when programming. 
The available entry points into GenX is:
<dl>
    <dt><b><code>frame</code></b></dt>
    <dd>The top level frame. All the widgets and windows of GenX can be acessed
    in this object</dd>
    <dt><b><code>data</code></b></dt>
    <dd>The data object. All the loaded data can acessed through this object.
    Each data set is acessed as in a list, i.e., data[0] for the first data set.
    x and y components are acesssed through data[0].x and data[0].y, 
    respectively.</dd>
    <dt><b><code>model</code></b></dt>
    <dd>Access to the model object. Stores the script and the module. 
    Important object.</dd>
</dl>
'''
from .. import add_on_framework as framework
import wx.py.shell, wx

class Plugin(framework.Template):
    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        inputpanel=self.NewInputFolder('Shell')
        sizer=wx.BoxSizer(wx.HORIZONTAL)
        inputpanel.SetSizer(sizer)
        self.shell=wx.py.shell.Shell(inputpanel, -1,
                                     locals={'frame': parent, 'model': self.GetModel(),
                                             'data': self.GetModel().get_data(),
                                             'ctrl': parent.model_control.controller})
        sizer.Add(self.shell, 1, wx.EXPAND)
        inputpanel.Layout()
        self.StatusMessage('Shell plugin loaded')



"""
Command classes that perform actions on a model. Allows
to implement undo/redo functionality and tracking of
actions in logs.
"""
import difflib

from logging import debug
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from enum import Flag, auto
from typing import List

from .model import Model
from .parameters import SortSplitItem
from .solver_basis import GenxOptimizer

class ModelInfluence(Flag):
    # The part of the model that gets altered by n action
    NONE = auto()
    SCRIPT = auto()
    PARAM = auto()
    DATA = auto()
    OPTIONS = auto()

class ModelAction(ABC):
    """
    Represents an action performed on the model.
    """
    model: Model

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def influences(self) -> ModelInfluence:
        ...

    @abstractmethod
    def __init__(self, model: Model, *params):
        ...

    @abstractmethod
    def execute(self):
        ...

    @abstractmethod
    def undo(self):
        ...

    def redo(self):
        self.execute()

    def __str__(self):
        # generate string representation replacing format string items interactively
        obj_items=self.__dict__.copy()
        obj_items.update(self.__class__.__dict__)
        try:
            return self.description.format(**obj_items)
        except KeyError:
            return self.description

    @property
    def description(self):
        return self.name

    @property
    def action_name(self):
        return self.name.format(**self.__dict__)

    def should_update(self, other: 'ModelAction'):
        # some actions prefer an update over keeping all changes in the todo
        return False

class NoOp(ModelAction):
    influences = ModelInfluence.NONE
    name = 'no action'
    def __init__(self, model): self.model = model
    def execute(self): pass
    def undo(self): pass
    def redo(self): pass

@dataclass
class ActionHistory:
    undo_stack: List[ModelAction] = field(default_factory=list)
    redo_stack: List[ModelAction] = field(default_factory=list)
    max_stack=100

    def execute(self, action: ModelAction):
        action.execute()
        action_string=str(action)
        action_string=action_string.replace('\n', '\n    ')
        debug(f'Action Executed - {action.__class__.__name__}: {action_string}')
        if len(self.undo_stack)>0 and action.should_update(self.undo_stack[-1]):
            self.undo_stack.pop()
        self.undo_stack.append(action)
        if len(self.undo_stack)>self.max_stack:
            self.undo_stack.pop(0)
        self.redo_stack=[]

    def undo(self) -> ModelAction:
        if len(self.undo_stack)>0:
            action=self.undo_stack[-1]
            action.undo()
            self.undo_stack.pop()
            self.redo_stack.append(action)
            return action
        else:
            return NoOp(None)

    def redo(self) -> ModelAction:
        if len(self.redo_stack)>0:
            action=self.redo_stack[-1]
            action.redo()
            self.redo_stack.pop()
            self.undo_stack.append(action)
            return action
        else:
            return NoOp(None)

    def remove_actions(self, start, length=1):
        """
        Undo the last 'start' actions, skipe 'length'
        redo actions and then redo the residual actions.
        """
        current_undo=len(self.redo_stack)
        changed_actions=[]
        skipped_action=[]
        redone_action=[]
        if len(self.undo_stack)<start:
            raise ValueError(f"Only {len(self.undo_stack)} items on the stack to undo.")
        for i in range(start):
            changed_actions.append(self.undo())
        for i in range(length):
            skipped_action.append(self.redo_stack.pop())
        while len(self.redo_stack)>current_undo:
            # redo actions that where not on the redo stack before the start
            try:
                redone_action.append(self.redo())
            except Exception as e:
                # there was an error during a redo, hald the process and
                # revert to previous state
                for _ in reversed(redone_action):
                    self.undo()
                skipped_action.reverse()
                self.redo_stack+=skipped_action
                while len(self.redo_stack)>current_undo:
                    self.redo()
                raise e
        return ActionBlock(changed_actions[-1].model, changed_actions)

    def clear(self):
        self.undo_stack=[]
        self.redo_stack=[]

class ActionBlock(ModelAction):
    influences = ModelInfluence.NONE
    name = 'action block'

    def __init__(self, model, actions: List[ModelAction]):
        self.model=model
        for action in actions:
            self.influences=self.influences|action.influences
        self.actions=actions

    def execute(self):
        for i, action in enumerate(self.actions):
            try:
                action.execute()
            except Exception as e:
                for rev in reversed(self.actions[:i]):
                    rev.undo()
                raise e

    def undo(self):
        for action in reversed(self.actions):
            action.undo()

    @property
    def description(self):
        return '|'.join(map(str, self.actions))


class UpdateModelScript(ModelAction):
    influences = ModelInfluence.SCRIPT

    def __init__(self, model, text):
        """
        Generate a diff that can be applied to the current script
        to get the new one. This allows the action to be performed
        even if a previous change is undone.
        """
        self.model=model
        self.old_text=self.model.get_script().replace('\r\n', '\n').replace('\r', '\n')
        debug(f'UpdateModelScript Init\nold text:\n    {self.old_text!r}\nnew text:\n    {text!r}')
        self.diff=self.generate_diff(self.old_text, text.replace('\r\n', '\n').replace('\r', '\n'))

    def execute(self):
        # Replace model script with new text, script as new text (toggles)
        self.old_text=self.model.get_script().replace('\r\n', '\n').replace('\r', '\n')
        new_text=self.apply_diff(self.old_text)
        self.model.set_script(new_text)

    def generate_diff(self, old, new):
        """
        Generate a list of changes with their line numbers
        as it is not clear how to change a script if more
        """
        d=difflib.Differ()
        diff=list(d.compare(old.splitlines(keepends=True), new.splitlines(keepends=True)))
        debug(f'generate_diff:\n{diff}')
        cdiff=[]
        old_indx=0
        while len(diff)>0:
            ndiff=diff.pop(0)
            if ndiff.startswith(' '):
                old_indx += 1
                if len(diff)>0 and diff[0][0] == '+':
                    nextcdiff=[old_indx, [], []]
                    while len(diff)>0 and diff[0][0] == '+':
                        nextcdiff[2].append(diff.pop(0)[2:])
                    cdiff.append(nextcdiff)
            elif ndiff.startswith('-'):
                nextcdiff = [old_indx, [ndiff[2:]], []]
                old_indx+=1
                # collect old lines to remove
                while len(diff)>0 and diff[0][0] == '-':
                    ndiff = diff.pop(0)
                    nextcdiff[1].append(ndiff[2:])
                    old_indx+=1
                while len(diff)>0 and diff[0][0] in ['?', '+']:
                    if diff[0][0]=='?':
                        ndiff = diff.pop(0)
                    else:
                        # collect new lines to insert
                        nextcdiff[2].append(diff.pop(0)[2:])
                cdiff.append(nextcdiff)
            elif old_indx==0 and ndiff.startswith('+'):
                nextcdiff = [old_indx, [], [ndiff[2:]]]
                while len(diff)>0 and diff[0][0] == '+':
                    nextcdiff[2].append(diff.pop(0)[2:])
                cdiff.append(nextcdiff)
            else:
                debug(f'ignored diff line {old_indx}: {ndiff}')
        return cdiff

    def apply_diff(self, old):
        old_list=old.splitlines(keepends=True)
        new_list=[]
        current_index=0
        for idx, diff_from, diff_to in self.diff:
            if old_list[idx:idx+len(diff_from)]==diff_from:
                new_list+=old_list[current_index:idx]
                new_list+=diff_to
                current_index=idx+len(diff_from)
            else:
                raise ValueError(f"The script hase been modefined in a way "
                                 f"that this change can no longer be applied: "
                                 f"line {idx} was {diff_from} but now is {old_list[idx:idx+len(diff_from)]}.")
        new_list+=old_list[current_index:]
        return ''.join(new_list)

    def format_diff(self):
        output='\n'
        for idx, diff_from, diff_to in self.diff:
            output += f'line {idx}:\n    '
            output += '    '.join(['-'+ln for ln in diff_from])
            if len(diff_from)>0: output += '    '
            output += '    '.join(['+'+ln for ln in diff_to])
        return output[:-1].replace('\n', '\n    ')

    def undo(self):
        self.model.set_script(self.old_text)

    @property
    def name(self):
        name = 'edit script ('
        name += '|'.join([str(d[0]) for d in self.diff])
        name += ')'
        return name

    @property
    def description(self):
        return self.name+': '+self.format_diff()

class UpdateSolverOptoins(ModelAction):
    influences = ModelInfluence.OPTIONS
    name = 'optimizer options'
    description = 'set optimizer parameters: {new_values}'

    def __init__(self, model, optimizer: GenxOptimizer, new_values: dict):
        self.model=model
        self.optimizer=optimizer
        self.new_values=new_values.copy()
        self.old_values={}
        self.combined_options=self.model.solver_parameters|self.optimizer.opt
        for key in new_values.keys():
            self.old_values[key]=getattr(self.combined_options, key, None)

    def execute(self):
        self.combined_options=self.model.solver_parameters|self.optimizer.opt
        for key in self.new_values.keys():
            self.old_values[key]=getattr(self.combined_options, key, None)
        # Update parameters and write to config
        for key, value in self.new_values.items():
            setattr(self.combined_options, key, value)
        self.model.WriteConfig()
        self.optimizer.WriteConfig()

    def undo(self):
        # Reset previous settings
        for key, value in self.old_values.items():
            setattr(self.combined_options, key, value)
        self.model.WriteConfig()
        self.optimizer.WriteConfig()

class UpdateDataPlotSettings(ModelAction):
    influences = ModelInfluence.DATA
    name = 'plot settings'

    def __init__(self, model, indices: List[int], sim_par: dict, data_par: dict):
        self.model=model
        self.indices=indices
        self.new_sim_par=sim_par.copy()
        self.new_data_par=data_par.copy()
        self.old_sim_pars={}
        self.old_data_pars={}

    def execute(self):
        # Update parameters for each dataset index
        for index in self.indices:
            di=self.model.data[index]
            self.old_sim_pars[index]=di.get_sim_plot_items()
            self.old_data_pars[index]=di.get_data_plot_items()
            di.set_sim_plot_items(self.new_sim_par)
            di.set_data_plot_items(self.new_data_par)

    def undo(self):
        # Reset previous settings
        for index in self.indices:
            di=self.model.data[index]
            di.set_sim_plot_items(self.old_sim_pars[index])
            di.set_data_plot_items(self.old_data_pars[index])

class UpdateColorCycle(ModelAction):
    influences = ModelInfluence.DATA
    name = 'color cycle'

    def __init__(self, model, source):
        self.model=model
        self.new_source=source
        self.old_source=None
        self.old_sim_colors={}
        self.old_data_colors={}

    def execute(self):
        # Update colors from each dataset index
        for i, di in enumerate(self.model.data):
            self.old_sim_colors[i]=di.sim_color
            self.old_data_colors[i]=di.data_color
        self.old_source=self.model.data.color_source
        self.model.data.update_color_cycle(self.new_source)

    def undo(self):
        # Reset previous settings and colors
        self.model.data.update_color_cycle(self.old_source)
        for i, di in enumerate(self.model.data):
            di.sim_color=self.old_sim_colors[i]
            di.data_color=self.old_data_colors[i]

class UpdateParams(ModelAction):
    influences = ModelInfluence.PARAM
    name = 'parameter values'
    description = 'set parameter values\n{param_names}\nfrom\n{old_values}\nto\n{new_values}'

    def __init__(self, model, new_values):
        self.model=model
        self.new_values=list(new_values)
        self.param_names=self.model.parameters.get_fit_pars()[1]
        self.old_values=self.model.parameters.get_value_pars()

    def execute(self):
        param_names=self.model.parameters.get_fit_pars()[1]
        if self.param_names!=param_names:
            raise ValueError("Current fit parameters do not correspond to setting during action init")
        self.old_values=self.model.parameters.get_value_pars()
        self.model.parameters.set_value_pars(self.new_values)

    def undo(self):
        self.model.parameters.set_value_pars(self.old_values)

class UpdateParamValue(ModelAction):
    influences = ModelInfluence.PARAM
    name = 'update parameter [{param_name}]'
    description = 'parameter[{param_name},{col}]: {old_value!r} -> {new_value!r}'

    def __init__(self, model, row, col, new_value):
        self.model=model
        self.col=col
        self.new_value=new_value
        if col!=0:
            self.param_name=self.model.parameters.get_names()[row]
        else:
            self.param_name=row
        self.old_value=self.model.parameters.get_value(row, col)

    def execute(self):
        if self.col!=0:
            row=self.model.parameters.get_names().index(self.param_name)
        else:
            row=self.param_name
        self.old_value=self.model.parameters.get_value(row, self.col)
        self.model.parameters.set_value(row, self.col, self.new_value)

    def undo(self):
        if self.col!=0:
            row=self.model.parameters.get_names().index(self.param_name)
        else:
            row=self.param_name
        self.model.parameters.set_value(row, self.col, self.old_value)

    def should_update(self, other):
        # if the same parameter is changed multiple time, just keep one change
        if type(other)!=type(self) or self.param_name!=getattr(other, 'param_name', None):
            return False
        else:
            self.old_value=other.old_value
            return True

class MoveParam(ModelAction):
    influences = ModelInfluence.PARAM
    name = 'move parameter'

    def __init__(self, model, row:int, step:int):
        self.model=model
        self.step=step
        name=self.model.parameters.get_names()[row]
        if name.strip()=='':
            self.param_name=row
        else:
            self.param_name=name

    def execute(self):
        if type(self.param_name) is int:
            row=self.param_name
        else:
            row=self.model.parameters.get_names().index(self.param_name)
        self.model.parameters.move_row(row, self.step)

    def undo(self):
        if type(self.param_name) is int:
            row=self.param_name+self.step
        else:
            row=self.model.parameters.get_names().index(self.param_name)
        self.model.parameters.move_row(row, -self.step)

class DeleteParams(ModelAction):
    influences = ModelInfluence.PARAM
    name = 'delete parameters'
    description = 'deleted {param_names}'

    def __init__(self, model, rows):
        self.model=model
        names=self.model.parameters.get_names()
        self.param_names=[names[ri] for ri in rows if names[ri].strip()!='']
        self.empty_rows=[ri for ri in rows if names[ri].strip()=='']
        self.old_data=[]

    def execute(self):
        names=self.model.parameters.get_names()
        rows=[names.index(ni) for ni in self.param_names]
        # make sure we only delete empty rows if they are still empty
        rows+=[row for row in self.empty_rows if names[row].strip()=='']
        old_rows=self.model.parameters.data
        self.old_data=[(ri, old_rows[ri]) for ri in rows]
        self.model.parameters.delete_rows(rows)

    def undo(self):
        for ri, di in self.old_data:
            self.model.parameters.data.insert(ri, di)

class InsertParam(ModelAction):
    influences = ModelInfluence.PARAM
    name = 'insert parameter'

    def __init__(self, model, row):
        self.model=model
        if row>=self.model.parameters.get_len_rows():
            self.param_name=self.model.parameters.get_len_rows()
        else:
            name=self.model.parameters.get_names()[row]
            if name.strip()=='':
                self.param_name=row
            else:
                self.param_name=name

    def execute(self):
        if type(self.param_name) is int:
            row=self.param_name
        else:
            row=self.model.parameters.get_names().index(self.param_name)
        self.insert_index=row
        self.model.parameters.insert_row(row)

    def undo(self):
        self.model.parameters.delete_rows([self.insert_index])

class SortAndGroupParams(ModelAction):
    influences = ModelInfluence.PARAM
    name = 'sort parameters'
    description = 'sort by {sort_params}'

    def __init__(self, model, sort_params: SortSplitItem):
        self.model=model
        self.sort_params=sort_params
        self.old_parameters=[]

    def execute(self):
        self.old_parameters=self.model.parameters.data.copy()
        if not self.model.compiled: self.model.compile_script()
        self.model.parameters.sort_rows(self.model, self.sort_params)
        self.model.parameters.group_rows(self.model, self.sort_params)
        self.model.parameters.strip()

    def undo(self):
        self.model.parameters.data=self.old_parameters
        self.old_parameters=[]

class ExchangeModel(ModelAction):
    influences = ModelInfluence.PARAM|ModelInfluence.DATA|ModelInfluence.SCRIPT
    name = 'exchange model'
    description = 'replace model, data and parameter grid from internal storage memory'

    def __init__(self, model, new_model: Model):
        self.model = model
        self.new_state = new_model.__getstate__().copy()
        self.new_state['filename'] = self.model.filename
        self.old_state = model.__getstate__().copy()

    def execute(self):
        self.model.__setstate__(self.new_state)

    def undo(self):
        self.model.__setstate__(self.old_state)



"""
Messaging protocol for remote model refinement with GenX server.
"""

import zlib
import struct
import asyncio
from abc import ABC
from dataclasses import dataclass
from enum import Enum
from logging import debug
from pickle import dumps, loads
from typing import Union

from ..diffev import DiffEvConfig
from ..model import Model
from ..solver_basis import GenxOptimizer, SolverParameterInfo, SolverResultInfo, SolverUpdateInfo


header_types = {}


@dataclass
class GenXMessage(ABC):
    """
    Base class for all messages. Does handle reading the
    header strings and unpacking the objects from the message body.

    The message classes are data classes with arbitray information
    as fields that all get send. Any other methods may be provided
    to work with the data.

    The Constructed message has the following form:
    ## Header Length ## (4 bytes, single unsigned integer)
    ## Header Type ## (4 bytes, single unsigned integer) Defines the message class for the header
    ## Header Data ## (pickled header with length {Header Length}-24)
    ## Data Length ## (8 bytes, single unsigned long long)
    ## Data ## (bz2 compressed pickled class with data)
    """
    HEADER_TYPE = 0

    @classmethod
    def use_header(cls, header_data):
        # has to be overwritten if HEADER_TYPE !+ 0
        pass

    def make_header(self) -> bytes:
        """
        Messages do not have to provide a header but it may be
        usefull if information about the data should be used
        while it is transmitted.
        """
        return b''

    def make_data(self):
        return dumps(self)

    def message(self):
        header = self.make_header()
        data = self.make_data()
        data = zlib.compress(data, 1)
        message = struct.pack("I", len(header)+4)
        message += struct.pack("I", self.HEADER_TYPE)
        message += header
        message += struct.pack("I", len(data))
        message += data
        return message

    @staticmethod
    async def receive(io: asyncio.StreamReader) -> "GenXMessage":
        debug('GenXMessage receiving')
        header_length = struct.unpack("I", await io.read(4))[0]
        header_type = struct.unpack("I", await io.read(4))[0]
        debug(f'GenXMessage header_length={header_length} ; header_type={header_type}')
        if header_type!=0:
            header_string = await io.read(header_length)
            header_data = loads(header_string)
            header_types[header_type].use_header(header_data)
        data_length = struct.unpack("I", await io.read(4))[0]
        debug(f'GenXMessage data_length={data_length}')
        data_string = await io.read(data_length)
        debug(f'Length of data_string received: {len(data_string)}')
        while len(data_string)<data_length:
            data_string += await io.read(data_length-len(data_string))
        data_string = zlib.decompress(data_string)
        res = loads(data_string)
        debug(f'GenXMessage of type {type(res)} sucessfully unpacked')
        return res


@dataclass
class StingMessage(GenXMessage):
    text: str


@dataclass
class EchoMessage(GenXMessage):
    # A string message that is send back by the server.
    text: str


class ActionType(int, Enum):
    START_FIT = 1
    STOP_FIT = 2


@dataclass
class ActionMessage(GenXMessage):
    action_type: ActionType
    short_info: str
    description: str


@dataclass
class ModelTransfer(GenXMessage):
    model: Model
    fitparams: DiffEvConfig


@dataclass
class OptimizerUpdate(GenXMessage):
    payload: Union[SolverUpdateInfo, SolverParameterInfo, SolverResultInfo]

    def __repr__(self):
        return f'OptimizerUpdate(payload={self.payload.__class__.__name__})'



'''
Use Levenberg-Marquardt as minimizer to estimate errors and for fast decent to next local minimum.
'''
import pickle
import _thread
from dataclasses import dataclass

from numpy import *
from scipy.optimize import leastsq

from .exceptions import ErrorBarError, OptimizerInterrupted
from .core.config import BaseConfig
from .core.custom_logging import iprint
from .model import Model
from .solver_basis import GenxOptimizer, GenxOptimizerCallback, SolverParameterInfo, SolverResultInfo, SolverUpdateInfo

class LMDefaultCallbacks(GenxOptimizerCallback):

    def text_output(self, text):
        iprint(text)
        sys.stdout.flush()

    def plot_output(self, update_data):
        pass

    def parameter_output(self, param_info):
        pass

    def fitting_ended(self, result_data):
        pass

    def autosave(self):
        pass

@dataclass
class LMConfig(BaseConfig):
    section='solver'

    groups={}

class LMOptimizer(GenxOptimizer):
    '''
    Optimizer based on Levenberg-Marquardt algorithm.
    '''
    opt: LMConfig
    model: Model
    fom_log: ndarray
    start_guess: ndarray
    covar: ndarray

    _callbacks: GenxOptimizerCallback=LMDefaultCallbacks()

    n_fom_evals=0

    def is_running(self):
        return False

    def __init__(self):
        GenxOptimizer.__init__(self)
        self.model=Model()
        self.fom_log=array([[0, 0]])[0:0]
        self.covar=None

    def pickle_string(self, clear_evals: bool = False):
        return pickle.dumps(self)

    def pickle_load(self, pickled_string: bytes):
        obj=pickle.loads(pickled_string, encoding='latin1', errors='ignore')
        # TODO: set own options from object

    def get_start_guess(self):
        return self.start_guess

    def get_model(self) -> Model:
        return self.model

    def get_fom_log(self):
        return self.fom_log

    def connect_model(self, model_obj: Model):
        '''
        Connects the model [model] to this object. Retrives the function
        that sets the variables  and stores a reference to the model.
        '''
        # Retrieve parameters from the model
        (param_funcs, start_guess, par_min, par_max)=model_obj.get_fit_pars(use_bounds=False)

        # Control parameter setup
        self.par_funcs=param_funcs
        self.model=model_obj
        self.n_dim=len(param_funcs)
        self.start_guess=start_guess

    def calc_sim(self, vec):
        ''' calc_sim(self, vec) --> None
        Function that will evaluate the the data points for
        parameters in vec.
        '''
        model_obj=self.model
        # Set the parameter values
        list(map(lambda func, value: func(value), self.par_funcs, vec))

        self.model.evaluate_sim_func()
        return self.model.fom

    def calc_fom(self, vec):
        '''
        Function to calcuate the figure of merit for parameter vector
        vec.
        '''
        if self._stop_fit:
            raise OptimizerInterrupted("interrupted")
        # Set the parameter values
        list(map(lambda func, value: func(value), self.par_funcs, vec))
        fom=self.model.evaluate_fit_func(get_elements=True) # fom is squared in leastsq
        chi=sign(fom)*sqrt(abs(fom))
        self.n_fom_evals+=1
        return chi

    def calc_error_bar(self, index: int) -> (float, float):
        if self.covar is None:
            raise ErrorBarError("Could not get covariance matrix from fit, maybe the parameters are coupled/have no influence?")
        err=sqrt(self.covar[index, index])
        return -err, err

    def project_evals(self, index: int):
        # -> (ArrayLike, ArrayLike)
        pass

    def start_fit(self, model: Model):
        self.n_fom_evals=0
        self.connect_model(model)
        self._stop_fit=False
        _thread.start_new_thread(self.optimize, ())

    def optimize(self):
        try:
            res=leastsq(self.calc_fom, self.start_guess, full_output=True)
        except OptimizerInterrupted:
            self._callbacks.fitting_ended(self.get_result_info(interrupted=True))
            return
        self.best_vec=res[0]
        if res[1] is None:
            self.covar=None
        else:
            Chi2Res = self.calc_fom(self.best_vec)**2
            s_sq = Chi2Res.sum()/(len(Chi2Res)-len(res[0]))  # variance of the residuals
            self.covar=res[1]*s_sq

        self.plot_output()
        self._callbacks.fitting_ended(self.get_result_info())

    def stop_fit(self):
        self._stop_fit=True

    def resume_fit(self, model: Model):
        pass

    def is_fitted(self):
        return self.n_fom_evals>0

    def is_configured(self) -> bool:
        pass

    def set_callbacks(self, callbacks: GenxOptimizerCallback):
        self._callbacks=callbacks

    def get_callbacks(self) -> 'GenxOptimizerCallback':
        return self._callbacks

    def plot_output(self):
        self.calc_sim(self.best_vec)
        data=SolverUpdateInfo(
            fom_value=self.model.fom,
            fom_name=self.model.fom_func.__name__,
            fom_log=self.get_fom_log(),
            new_best=True,
            data=self.model.data
            )
        self._callbacks.plot_output(data)

    def get_result_info(self, interrupted=False):
        result = SolverResultInfo(
            start_guess=self.start_guess.copy(),
            error_message="",
            values=self.best_vec.copy(),
            new_best=not interrupted,
            population=[],
            max_val=[],
            min_val=[],
            fitting=False
            )
        return result




'''
Configuration and config file handling.
'''
import dataclasses
import io
from configparser import ConfigParser
from functools import lru_cache
from abc import ABC, abstractmethod
from typing import Type, get_type_hints, Dict, List
from logging import debug

from .custom_logging import iprint
from ..exceptions import GenxIOError, GenxOptionError

# ==============================================================================
class Config:
    """
    GenX configuration handler.
    """
    def __init__(self):
        self.default_config=ConfigParser()
        self.model_config=ConfigParser()

    def load_default(self, filename, reset=False):
        '''
        Loads the default config from file filename. Raises a IOError if the
        can not be found.
        '''
        try:
            self.default_config.read(filename)
        except Exception as e:
            iprint(e)
            raise GenxIOError(f'Could not load default config file:\n {e}', filename)
        if reset:
            self.model_config=ConfigParser()

    def write_default(self, filename):
        '''
        Writes the current default configuration to filename
        '''
        try:
            with open(filename, 'w') as cfile:
                self.default_config.write(cfile)
        except Exception as e:
            iprint(e)
            raise GenxIOError(f'Could not write default config file:\n {e}', filename)

    def load_string(self, string: str):
        '''
        Loads a config from a string input.  Raises an IOError if the string can not be
        read.
        '''
        self.model_config=ConfigParser()
        try:
            self.model_config.read_string(string)
        except Exception as e:
            raise GenxIOError(f'Could not load model config file:\n {e}')

    def _getf(self, method: str, section: str, option: str, fallback=None):
        '''
        For the function function try to locate the section and option first in
        model_config if that fails try to locate it in default_config. If both
        fails raise a GenxOptionError.
        '''
        if self.model_config.has_option(section, option):
            func=getattr(self.model_config, method)
        elif self.default_config.has_option(section, option):
            func=getattr(self.default_config, method)
        else:
            raise GenxOptionError(section, option)
        try:
            return func(section, option, fallback=fallback)
        except ValueError:
            if fallback is None:
                raise GenxOptionError(section, option)
            else:
                debug(f'Return fallback due to ValueError in _getf, section={section}/option={option}')
                return fallback

    def getlist(self, section: str, option: str, fallback=None):
        str_value=self.get(section, option, fallback=fallback)
        return [s.strip() for s in str_value.split(';')]

    @lru_cache(maxsize=10)
    def __getattr__(self, item: str):
        """
        Handle all get functions from ConfigParser to be redirected through _getf,
        everything else is treated normally.
        The lambda functions are cached so they only get generated once.
        """
        if item.startswith('get') and hasattr(self.model_config, item):
            return lambda section, option, fallback=None: self._getf(item, section, option, fallback=fallback)
        else:
            raise AttributeError(f"'Config' object has no attribute '{item}'")

    def model_set(self, section, option, value):
        '''
        Set a value in section, option for the model configuration.
        '''
        if not self.model_config.has_section(section):
            self.model_config.add_section(section)
        self.model_config.set(section, option, str(value))

    def default_set(self, section, option, value):
        '''
        Set a value in section, option for the model configuration.
        '''
        if not self.default_config.has_section(section):
            self.default_config.add_section(section)
        self.default_config.set(section, option, str(value))

    def set(self, section, option, value):
        '''
        Set a value in section, option for the model configuration. 
        Just a duplication of model_set.
        '''
        self.model_set(section, option, value)

    def model_dump(self):
        '''
        dumps the model configuration to a string.
        '''
        buffer=io.StringIO()
        self.model_config.write(buffer)
        string=buffer.getvalue()
        buffer.close()
        return string

# create the global GenX config object used by the program, it does not depend on anything else
config=Config()

class BaseConfig(ABC):
    """
    Base class for handling configuration that shall be stored by a GenX Config object.

    Handles general parameter read/write so that derived classes only have to
    inherit from it and use dataclass attributes.
    """
    section:str

    @property
    @abstractmethod
    def section(self):
        raise NotImplementedError("Subclass needs to define the section attribute.")

    def asdict(self):
        # noinspection PyDataclass
        return dataclasses.asdict(self)

    def load_config(self):
        field: dataclasses.Field
        # noinspection PyDataclass
        for field in dataclasses.fields(self):
            if field.type is float:
                getf=config.getfloat
            elif field.type is int:
                getf=config.getint
            elif field.type is bool:
                getf=config.getboolean
            elif field.type is list:
                getf=config.getlist
            else:
                getf=config.get

            if field.default is dataclasses.MISSING:
                fallback=None
            else:
                fallback=field.default
            option=field.name.replace('_', ' ')
            try:
                value=getf(self.section, option, fallback=fallback)
            except GenxOptionError as e:
                debug(f'Could not read option {self.section}/{option} of type {field.type.__name__}:\n    {e}')
            else:
                setattr(self, field.name, value)

    def safe_config(self, default=False):
        data=self.asdict()
        if default:
            setter=config.default_set
        else:
            setter=config.model_set
        for key, value in data.items():
            option=key.replace('_', ' ')
            if type(value) is list:
                value=';'.join(value)
            setter(self.section, option, value)

    def copy(self):
        # noinspection PyDataclass
        return dataclasses.replace(self)

    @property
    def groups(self)->Dict[str,List[str]]:
        """
        Can be replaced by dictionary in sub-classes to define how to group parameters in display and
        configuration dialog entries.
        If a list entry contains a list of strings it is used to group the values horizontally.
        A list of 4 items with the second beeing a boolean is interpreted as a choice between two parameters:
            [{Choice Group Label}, bool-first is active, first value, second value]
        """
        res=list(self.asdict().keys())
        return {'': res}

    def get_fields(self, fltr=None)->List[dataclasses.Field]:
        """
        Return the fields for the dataclass.
        If fltr is supplied the results will be filtered by field names given in this list.
        """
        # noinspection PyDataclass
        fields=dataclasses.fields(self) # child classes will be dataclass objects
        if fltr is None:
            return list(fields)
        else:
            out=[]
            for fi in fields:
                if fi.name in fltr:
                    out.append(fi)
            return out

    @staticmethod
    def GParam(default, pmin=None, pmax=None, label=None, descriptoin=None):
        # allow to add some metadata to the parameter to use in dialog generation
        gmeta={}
        if pmin is not None:
            gmeta['pmin']=pmin
        if pmax is not None:
            gmeta['pmax']=pmax
        if label is not None:
            gmeta['label']=label
        if descriptoin is not None:
            gmeta['descriptoin']=descriptoin
        return dataclasses.field(default=default, metadata={'genx': gmeta})

    @staticmethod
    def GChoice(default, selection, label=None, descriptoin=None):
        """
        Allow to add some metadata to the parameter to use in dialog generation,
        selection is a list of alloweed string values for this parameter.
        """
        gmeta={'selection': selection}
        if label is not None:
            gmeta['label']=label
        if descriptoin is not None:
            gmeta['descriptoin']=descriptoin
        return dataclasses.field(default=default, metadata={'genx': gmeta})

    def __or__(self, other):
        return MergedConfig(self, other)

class MergedConfig(BaseConfig):
    """
    A config-like object that combines two or more config objects.

    Attribut access is passed to the child objects.
    """
    section = None
    _children: List[BaseConfig]

    def __init__(self, *children):
        self._children=list(children)

    def __or__(self, other):
        all_children=self._children+[other]
        return MergedConfig(*all_children)

    def get_fields(self, fltr=None) ->List[dataclasses.Field]:
        fout=[]
        for c in self._children:
            fout+=c.get_fields(fltr)
        return fout

    def copy(self):
        clist=[c.copy() for c in self._children]
        return MergedConfig(*clist)

    @property
    def groups(self) ->Dict[str,List[str]]:
        dout={}
        for c in self._children:
            dout.update(c.groups)
        return dout

    def __getattr__(self, name):
        # search through the child configuration fields if name is in there and return value
        for c in self._children:
            fnames=[fi.name for fi in c.get_fields()]
            if name in fnames:
                return getattr(c, name)
        raise AttributeError("{item} not found in any child configuration")

    def __setattr__(self, name, value):
        if name.startswith('_'):
            BaseConfig.__setattr__(self, name, value)
        # search through the child configuration fields if name is in there, otherwise set own attribute
        for c in self._children:
            fnames=[fi.name for fi in c.get_fields()]
            if name in fnames:
                return setattr(c, name, value)
        BaseConfig.__setattr__(self, name, value)


class Configurable:
    """
    A mixin class for classes that store parameters in a savable configuration.

    Defines methods for reading and writing configurations. Subclasses can
    define the config_updated method to be called after reading a configuration.

    If no config_class is passed to the constructor explicitly, any existing type hint
    to a BaseConfig derived object will be used to configure the opt attribute.
    """
    def __init__(self, config_class: Type[BaseConfig]=None):
        if config_class is None:
            hints=get_type_hints(self)
            config_class=hints.get('opt', None)

        if not issubclass(config_class, BaseConfig):
            raise ValueError("Configurable needs either an explicit config_class "
                             "or 'opt' type hint derived from BaseConfig")

        self.opt=config_class()

    def ReadConfig(self):
        """ Reads the variables stored in the config file."""
        self.opt.load_config()
        self.UpdateConfigValues()

    def WriteConfig(self, default=False):
        """Writes the varaibles to be stored to the config"""
        self.opt.safe_config(default=default)

    def UpdateConfigValues(self):
        """
        Sub-classes can overwrite this method to preform some action after
        a configuration has been loaded.
        """



"""
Typing support for reflectivity plugins.
Could be extended with other small helpers.
"""
from genx.data import DataList
from genx.models.lib.refl import InstrumentBase, LayerBase, SampleBase, StackBase


try:
    # noinspection PyUnresolvedReferences
    from typing import Protocol, Iterable, List, Dict, Any, Type, Union
except ImportError:
    ReflectivityModule = object
else:
    class Sample(Protocol):

        def SimSLD(self, z, item: Union[str, None], inst: InstrumentBase) -> Dict[str, Any]:
            ...

    def Sim(data:DataList) -> List[Any]:
        ...

    class ReflectivityModule(Protocol):
        """
        Defines the attributes that are expected in a python module that defines a reflectivity model.
        This is used to type check what the classes in this file use.
        """
        sample_string_choices: Dict[str, Iterable[str]]
        SampleParameters: Dict[str, Any]
        SampleGroups: Iterable
        SampleUnits: Dict[str, str]

        instrument_string_choices: Dict[str, Iterable[str]]
        InstrumentParameters: Dict[str, Any]
        InstrumentGroups: Iterable
        InstrumentUnits: Dict[str, str]

        LayerParameters: Dict[str, Any]
        LayerGroups: Iterable
        LayerUnits: Dict[str, str]

        StackParameters: Dict[str, Any]
        StackGroups: Iterable
        StackUnits: Dict[str, str]

        Instrument: Type[InstrumentBase]
        Sample: Type[SampleBase]
        Stack: Type[StackBase]
        Layer: Type[LayerBase]

        _sim: bool
        SLD: List[Dict[str, Any]]
        sample: Sample
        Sim: Sim
        inst: InstrumentBase



'''
Library that contains the the class Model. This
is a general class that store and binds togheter all the other
classes that is part of model parameters, such as data and parameters.
'''

import inspect
import os
import pickle as pickle
import traceback
import types
import typing
import zipfile
from dataclasses import dataclass, field
from io import StringIO
from logging import debug, info
from copy import deepcopy

import numpy as np

# GenX libraries
from . import fom_funcs
from .core.config import BaseConfig
from .core.custom_logging import iprint
from .core.h5_support import H5HintedExport
from .data import DataList
from .exceptions import FomError, GenxIOError, ModelError, ParameterError
from .models.lib.parameters import get_parameters, NumericParameter
from .parameters import Parameters


@dataclass
class StartupScript(BaseConfig):
    section = 'startup'
    script: str = '[""]'


@dataclass
class ModelParameters(BaseConfig):
    section = 'parameters'
    registred_classes: list = field(default_factory=list)
    set_func: str = 'set'


@dataclass
class SolverParameters(BaseConfig):
    section = 'solver'
    figure_of_merit: str = BaseConfig.GChoice('log', selection=fom_funcs.func_names)
    ignore_fom_nan: bool = True
    ignore_fom_inf: bool = True

    limit_fit_range: bool = False
    fit_xmin: float = BaseConfig.GParam(0.0, pmin=-1000., pmax=1000.)
    fit_xmax: float = BaseConfig.GParam(180.0, pmin=-1000., pmax=1000.)

    groups = {
        'FOM': ['figure_of_merit', ['ignore_fom_nan', 'ignore_fom_inf'],
                'limit_fit_range', ['fit_xmin', 'fit_xmax']]
        }


class GenxScriptModule(types.ModuleType):
    data: DataList
    _sim: bool

    def __init__(self, data: DataList):
        types.ModuleType.__init__(self, 'genx_script_module')
        self.__package__ = fom_funcs.__package__
        self.data = data
        self._sim = False
        self.TextIO = typing.TextIO

    @staticmethod
    def Sim(data: DataList):
        # default implementation of Sim function, will be overwritten by user.
        return [di.y*0 for di in data]


class Model(H5HintedExport):
    ''' A class that holds the model i.e. the script that defines
        the model and the data + various other attributes.
    '''
    h5group_name = 'current'

    saved = True
    fom = None
    fom_mask_func = None

    # parameters stored to file
    script: str
    fomfunction: str
    data: DataList
    parameters: Parameters
    sequence_value: float = 0.

    @property
    def fomfunction(self):
        return self.fom_func.__name__

    @fomfunction.setter
    def fomfunction(self, value):
        if value in fom_funcs.func_names:
            self.set_fom_func(eval('fom_funcs.'+value))
            self.solver_parameters.figure_of_merit = value
        else:
            iprint("Can not find fom function name %s"%value)

    def __init__(self):
        '''
        Create a instance and init all the variables.
        '''
        self.opt = ModelParameters()
        self.startup_script = StartupScript()
        self.solver_parameters = SolverParameters()
        self.ReadConfig()

        self.data = DataList()
        self.set_script('')
        # noinspection PyBroadException
        try:
            self.set_script("\n".join(eval(self.startup_script.script)))
        except:
            debug('Issue when loading script from config:', exc_info=True)
        self.parameters = Parameters()
        self.fom_func = fom_funcs.log

        self._reset_module()

        # Temporary stuff that needs to keep track on
        self.filename = ''

        self.extra_analysis = {}

    def ReadConfig(self):
        self.opt.load_config()
        self.startup_script.load_config()
        self.solver_parameters.load_config()
        self.create_fom_mask_func()
        self.set_fom_from_config()

    def WriteConfig(self):
        self.opt.safe_config()
        self.startup_script.safe_config()
        self.solver_parameters.safe_config()
        self.set_fom_from_config()

    def set_fom_from_config(self):
        self.fomfunction = self.solver_parameters.figure_of_merit

    def load(self, filename):
        ''' 
        Function to load the necessary parameters from a model file.
        '''
        try:
            loadfile = zipfile.ZipFile(filename, 'r')
        except Exception as e:
            raise GenxIOError(f'Could not open file, python error {e!r}', filename)
        try:
            new_data = pickle.loads(loadfile.read('data'), encoding='latin1', errors='ignore')
            self.data.safe_copy(new_data)
        except Exception as e:
            iprint('Data section loading (gx file) error:\n ', e, '\n')
            raise GenxIOError('Could not locate the data section.', filename)
        try:
            self.set_script(pickle.loads(loadfile.read('script'), encoding='latin1', errors='ignore'))
        except Exception as e:
            iprint('Script section loading (gx file) error:\n ', e, '\n')
            raise GenxIOError('Could not locate the script.', filename)

        try:
            new_parameters = pickle.loads(loadfile.read('parameters'), encoding='latin1', errors='ignore')
            self.parameters.safe_copy(new_parameters)
        except Exception as e:
            iprint('Script section loading (gx file) error:\n ', e, '\n')
            raise GenxIOError('Could not locate the parameters section.', filename)
        try:
            self.fom_func = pickle.loads(loadfile.read('fomfunction'), encoding='latin1', errors='ignore')
        except Exception:
            raise GenxIOError('Could not locate the fomfunction section.', filename)

        loadfile.close()

        self.filename = os.path.abspath(filename)
        self.saved = True
        self.script_module = GenxScriptModule(self.data)

    def save(self, filename):
        '''
        Function to save the model to file filename
        '''
        try:
            savefile = zipfile.ZipFile(filename, 'w')
        except Exception as e:
            raise GenxIOError(str(e), filename)

        # Save the data structures to file
        try:
            savefile.writestr('data', pickle.dumps(self.data))
        except Exception as e:
            raise GenxIOError('Error writing data: '+str(e), filename)
        try:
            savefile.writestr('script', pickle.dumps(self.script))
        except Exception as e:
            raise GenxIOError('Error writing script: '+str(e), filename)
        self.parameters.model = None
        try:
            savefile.writestr('parameters', pickle.dumps(self.parameters))
        except Exception as e:
            raise GenxIOError('Error writing parameters: '+str(e), filename)
        try:
            savefile.writestr('fomfunction', pickle.dumps(self.fom_func))
        except Exception as e:
            raise GenxIOError('Error writing fom_func:  '+str(e), filename)

        savefile.close()

        self.filename = os.path.abspath(filename)
        self.saved = True

    def __getstate__(self):
        # generate a pickleable object for thie model, it cannot contain dynamically generated functions
        state = self.__dict__.copy()
        if 'fom_mask_func' in state:
            del state['fom_mask_func']
        if 'script_module' in state:
            del state['script_module']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.create_fom_mask_func()
        if self.compiled:  # if the model was compiled before pickling, do it now
            self.compile_script()

    def copy(self) -> "Model":
        state = self.__getstate__().copy()
        output = Model()
        output.__setstate__(state)
        return output

    def deepcopy(self) -> "Model":
        return deepcopy(self)

    def read_h5group(self, group):
        """
        Read the parameters from a hdf5 group
        """
        H5HintedExport.read_h5group(self, group)
        self.create_fom_mask_func()
        self.saved = True
        self.script_module = GenxScriptModule(self.data)
        self.compiled = False

    def save_addition(self, name, text):
        '''
        Save additional text sub-file with name name/text to the current file.
        '''
        if self.filename=='':
            raise GenxIOError('File must be saved before new information is added', '')
        try:
            savefile = zipfile.ZipFile(self.filename, 'a')
        except Exception as e:
            raise GenxIOError(str(e), self.filename)

        # Check so the model data is not overwritten
        if name=='data' or name=='script' or name=='parameters':
            raise GenxIOError('It not alllowed to save a subfile with name: %s'%name)

        try:
            savefile.writestr(name, text)
        except Exception as e:
            raise GenxIOError(str(e), self.filename)
        savefile.close()

    def load_addition(self, name) -> bytes:
        '''
        Load additional text from sub-file
        '''
        if self.filename=='':
            raise GenxIOError('File must be loaded before additional information is read', '')
        try:
            loadfile = zipfile.ZipFile(self.filename, 'r')
        except Exception:
            raise GenxIOError('Could not open the file', self.filename)

        try:
            raw_string = loadfile.read(name)
        except Exception:
            raise GenxIOError('Could not read the section named: %s'%name, self.filename)
        loadfile.close()
        return raw_string

    def reset(self):
        self._reset_module()

    def _reset_module(self):
        ''' 
        Internal method for resetting the module before compilation
        '''
        self.create_fom_mask_func()
        self.script_module = GenxScriptModule(self.data)
        self.compiled = False

    def compile_script(self):
        ''' 
        compile the script in a seperate module.
        '''
        self._reset_module()
        # Testing to see if this works under windows
        self.set_script('\n'.join(self.script.splitlines()))
        try:
            exec(self.script, self.script_module.__dict__)
        except Exception:
            outp = StringIO()
            traceback.print_exc(200, outp)
            val = outp.getvalue()
            outp.close()
            raise ModelError(str(val), 0)
        else:
            self.compiled = True

    def eval_in_model(self, codestring):
        '''
        Excecute the code in codestring in the namespace of model module
        '''
        result = eval(codestring, self.script_module.__dict__)
        return result

    def set_script_variable(self, name, value):
        self.script_module.__dict__[name] = value

    def unset_script_variable(self, name):
        del(self.script_module.__dict__[name])

    def create_fom_mask_func(self):
        """
        Create a mask func for fom to take care of unallowed values.
        """
        if self.solver_parameters.ignore_fom_nan and self.solver_parameters.ignore_fom_inf:
            fom_mask = lambda a: np.where(np.isfinite(a), a, np.zeros_like(a))
        elif self.solver_parameters.ignore_fom_nan:
            fom_mask = lambda a: np.where(np.isnan(a), a, np.zeros_like(a))
        elif self.solver_parameters.ignore_fom_inf:
            fom_mask = lambda a: np.where(np.isinf(a), a, np.zeros_like(a))
        else:
            fom_mask = lambda a: a

        self.fom_mask_func = fom_mask

    def calc_fom(self, simulated_data):
        '''
        Sums up the evaluation of the fom values calculated for each
        data point to form the overall fom function for all data sets.
        '''
        fom_raw = self.fom_func(simulated_data, self.data)
        # limit the x-range of fitting
        if self.solver_parameters.limit_fit_range:
            for i, di in enumerate(self.data):
                fltr = (di.x<self.solver_parameters.fit_xmin) | (di.x>self.solver_parameters.fit_xmax)
                fom_raw[i][fltr] = 0.
        # Sum up a unique fom for each data set in use
        fom_indiv = [np.sum(np.abs(self.fom_mask_func(fom_set))) for fom_set in fom_raw]
        fom = np.sum([f for f, d in zip(fom_indiv, self.data) if d.use])

        # Lets extract the number of data points as well:
        N = np.sum([len(self.fom_mask_func(fom_set)) for fom_set, d in zip(fom_raw, self.data) if d.use])
        # And the number of fit parameters
        p = self.parameters.get_len_fit_pars()
        # self.fom_dof = fom/((N-p)*1.0)
        try:
            use_dif = self.fom_func.__div_dof__
        except AttributeError:
            use_dif = False
        if use_dif:
            fom = fom/((N-p)*1.0)

        penalty_funcs = self.get_par_penalty()
        if len(penalty_funcs)>0 and fom is not np.NAN:
            fom += sum([pf() for pf in penalty_funcs])
        return fom_raw, fom_indiv, fom

    def evaluate_fit_func(self, get_elements=False):
        '''
        Evalute the Simulation fucntion and returns the fom. Use this one
        for fitting. Use evaluate_sim_func(self) for updating of plots
        and such.
        '''
        self.script_module._sim = False
        simulated_data = self.script_module.Sim(self.data)
        fom_raw, fom_inidv, fom = self.calc_fom(simulated_data)
        if get_elements:
            return np.hstack([self.fom_mask_func(fom_set) for fom_set in fom_raw])
        else:
            return fom

    def evaluate_sim_func(self):
        '''
        Evalute the Simulation function and updates the data simulated data
        as well as the fom of the model. Use this one for calculating data to
        update plots, simulations and such.
        '''
        self.script_module._sim = True
        try:
            simulated_data = self.script_module.Sim(self.data)
        except Exception:
            outp = StringIO()
            traceback.print_exc(200, outp)
            val = outp.getvalue()
            outp.close()
            raise ModelError(str(val), 1)

        # check so that the Sim function returns anything
        if not simulated_data:
            text = 'The Sim function does not return anything, it should'+ \
                   ' return a list of the same length as the number of data sets.'
            raise ModelError(text, 1)
        # Check so the number of data sets is correct
        if len(simulated_data)!=len(self.data):
            text = 'The number of simulated data sets returned by the Sim function' \
                   +' has to be same as the number of loaded data sets.\n'+ \
                   'Number of loaded data sets: '+str(len(self.data))+ \
                   '\nNumber of simulated data sets: '+str(len(simulated_data))
            raise ModelError(text, 1)

        self.data.set_simulated_data(simulated_data)

        try:
            fom_raw, fom_inidv, fom = self.calc_fom(simulated_data)
            self.fom = fom
        except Exception:
            outp = StringIO()
            traceback.print_exc(200, outp)
            val = outp.getvalue()
            outp.close()
            raise FomError(str(val))
        self.data.set_fom_data(fom_raw)

    def create_fit_func(self, identifier):
        '''
        Creates a function from the string expression in string. 
        If the string is a function in the model this function will be
        returned if string represents anything else a function that sets that 
        object will be returned.
        '''
        obj = self.eval_in_model(identifier)
        # Is it a function or a method!
        if callable(obj):
            return obj
        # Make a function to set the object
        elif isinstance(obj, NumericParameter):
            # We have a NumericParameter that should be set
            exec('def __tempfunc__(val):\n'
                 '    %s.value = val'%identifier, self.script_module.__dict__)
            # noinspection PyUnresolvedReferences
            return self.script_module.__tempfunc__
        else:
            # The function must be created in the module in order to access
            # the different variables
            exec('def __tempfunc__(val):\n\t%s = val'%identifier, self.script_module.__dict__)
            # noinspection PyUnresolvedReferences
            return self.script_module.__tempfunc__

    def get_fit_pars(self, use_bounds=True):
        '''
        Returns the parameters used with fitting. i.e. the function to 
        set the paraemters, the guess value (values), minimum allowed values
        and the maximum allowed values
        '''
        (row_numbers, sfuncs, vals, minvals, maxvals) = self.parameters.get_fit_pars()
        if len(sfuncs)==0:
            raise ParameterError(sfuncs, 0, None, 4)
        if use_bounds:
            # Check for min and max on all the values
            for i in range(len(vals)):
                # parameter less than min
                if vals[i]<minvals[i]:
                    raise ParameterError(sfuncs[i], row_numbers[i], None, 3)
                # parameter larger than max
                if vals[i]>maxvals[i]:
                    raise ParameterError(sfuncs[i], row_numbers[i], None, 2)

        # Compile the strings to create the functions..
        funcs = []
        for func in sfuncs:
            try:
                funcs.append(self.create_fit_func(func))
            except Exception as e:
                raise ParameterError(func, row_numbers[len(funcs)], e, 0)
        return funcs, vals, minvals, maxvals

    def get_par_penalty(self):
        for var in self.script_module.__dict__.values():
            if var.__class__.__name__=='UserVars':
                # noinspection PyProtectedMember
                return var._penalty_funcs
        return []

    def get_sim_pars(self):
        '''
        Returns the parameters used with simulations. i.e. the function to 
        set the parameters, the guess value (values). Used for simulation, 
        for fitting see get_fit_pars(self).s
        '''
        (sfuncs, vals) = self.parameters.get_sim_pars()
        # Compile the strings to create the functions..
        funcs = []
        for func in sfuncs:
            try:
                funcs.append(self.create_fit_func(func))
            except Exception as e:
                raise ParameterError(func, len(funcs), e, 0)

        return funcs, vals

    # noinspection PyShadowingBuiltins
    def simulate(self, compile=True):
        '''
        Simulates the data sets using the values given in parameters...
        also compiles the script if asked for (default)
        '''
        if compile:
            self.compile_script()
        (funcs, vals) = self.get_sim_pars()
        # Set the parameter values in the model
        i = 0
        for func, val in zip(funcs, vals):
            try:
                func(val)
            except Exception as e:
                (sfuncs_tmp, vals_tmp) = self.parameters.get_sim_pars()
                raise ParameterError(sfuncs_tmp[i], i, e, 1)
            i += 1

        self.evaluate_sim_func()

    def new_model(self):
        '''
        Reinitilizes the model. Thus, removes all the traces of the
        previous model. 
        '''
        debug("create new model")
        self.data = DataList()
        self.set_script('')
        self.parameters = Parameters()

        self.fom_func = fom_funcs.log
        self._reset_module()

        # Temporary stuff that needs to keep track on
        self.filename = ''
        self.saved = False

        self.extra_analysis = {}

    def pickable_copy(self):
        '''
        Creates a pickable object of the model. Can be used for saving or
        sending to other processes, i.e., parallel processing.
        '''
        model_copy = Model()
        model_copy.data = self.data
        model_copy.script = self.script
        model_copy.parameters = self.parameters.copy()
        model_copy.fom_func = self.fom_func
        # The most important stuff - a module is not pickleable
        model_copy.script_module = None
        model_copy.filename = self.filename
        model_copy.compiled = self.compiled
        model_copy.fom = self.fom
        model_copy.solver_parameters = self.solver_parameters.copy()
        model_copy.startup_script = self.startup_script.copy()
        model_copy.opt = self.opt.copy()
        model_copy.saved = self.saved
        # Needs to reset the fom_mask_func since this fails under windows.
        model_copy.fom_mask_func = None
        #
        return model_copy

    def get_table_as_ascii(self):
        '''
        Just a copy of the parameters class method get_ascii_output()
        '''
        return self.parameters.get_ascii_output()

    def get_data_as_asciitable(self, indices=None):
        '''
        Just a copy of the method defined in data with the same name.
        '''
        return self.data.get_data_as_asciitable(indices)

    def export_table(self, filename):
        '''
        Export the table to filename. ASCII output.
        '''
        self._save_to_file(filename, self.parameters.get_ascii_output())

    def export_orso(self, basename):
        '''
        Export the data to files with basename filename. ORT output.
        The fileending will be .ort
        '''
        self.simulate(True)
        from orsopy.fileio import save_orso, OrsoDataset, Orso, Column, ErrorColumn
        from .version import __version__ as version
        para_list = [dict([(nj, tpj(pij)) for nj, pij, tpj in zip(self.parameters.data_labels, pi,
                                                                  [str, float, bool, float, float, str])])
                     for pi in self.parameters.data if pi[0].strip()!='']
        add_header = {
            'analysis': {
                'software': {'name': 'GenX', 'version': version},
                'script': self.script,
                'parameters': para_list,
                }
            }
        # add possible additional analysis data to the header
        add_header['analysis'].update(self.extra_analysis)
        ds = []
        for di in self.data:
            header = Orso.empty().to_dict()
            self.update_dictionary(header, di.meta)
            try:
                import getpass
                # noinspection PyTypeChecker
                add_header['analysis']['operator'] = {'name': getpass.getuser()}
            except Exception:
                pass
            import datetime
            # noinspection PyTypeChecker
            add_header['analysis']['timestamp'] = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
            header.update(add_header)
            header['data_set'] = di.name
            columns = [di.x, di.y, di.error]
            if 'columns' in header:
                prev_columns_list = [Column(**ci) if 'name' in ci else ErrorColumn(**ci) for ci in header['columns']]
                prev_columns = dict([(ci.name, ci) for ci in prev_columns_list])
            else:
                prev_columns_list = None
                prev_columns = {}
            if prev_columns_list:
                column_names = prev_columns_list[:3]
            elif hasattr(self.script_module, 'inst') and self.script_module.inst.coords in ['tth', '2θ']:
                column_names = [Column('TTh', 'deg'), Column('R'), ErrorColumn('R')]
            else:
                column_names = [Column('Qz', '1/angstrom'), Column('R'), ErrorColumn('R')]
            if 'res' in di.extra_data:
                columns.append(di.extra_data['res'])
                column_names.append(ErrorColumn(column_names[0].name))
            columns.append(di.y_sim)
            column_names.append(Column('Rsim'))
            columns.append(di.y_fom)
            column_names.append(Column('FOM'))
            for name, col in sorted(di.extra_data.items()):
                if name=='res':
                    continue
                columns.append(col)
                if name in prev_columns:
                    column_names.append(prev_columns[name])
                else:
                    column_names.append(Column(name))
            header['columns'] = column_names
            ds.append(OrsoDataset(Orso(**header), np.array(columns).T))
        try:
            save_orso(ds, basename, data_separator='\n')
        except GenxIOError as e:
            raise GenxIOError(e.error_message, e.file)

    @staticmethod
    def update_orso_meta(datasets: DataList):
        from orsopy.fileio import Orso
        for di in datasets:
            defaults = Orso.empty().to_dict()
            Model.update_dictionary(defaults, di.meta)
            di.meta = defaults

    def validate_orso_meta(self):
        from orsopy.fileio import base
        base._validate_header_data([di.meta for di in self.data])

    @staticmethod
    def update_dictionary(to_update, to_include):
        """
        Updates the entries in one dictrionary recursively to keep existing sub-keys.
        """
        for key, value in to_include.items():
            if key in to_update and isinstance(to_update[key], dict):
                if not isinstance(value, dict):
                    info(f"Metadata for {key} according to specification should be dictionary, found {value}")
                else:
                    Model.update_dictionary(to_update[key], value)
            else:
                to_update[key]=value

    def export_data(self, basename):
        '''
        Export the data to files with basename filename. ASCII output. 
        The fileending will be .dat
        First column is the x-values. 
        Second column is the data y-vales.
        Third column the error on the data y-values.
        Fourth column the calculated y-values.
        '''
        try:
            self.data.export_data_to_files(basename)
        except GenxIOError as e:
            raise GenxIOError(e.error_message, e.file)

    def export_script(self, filename):
        '''
        Export the script to filename. Will be a python script with ASCII 
        output (naturally).
        '''
        self._save_to_file(filename, self.script)

    def import_script(self, filename):
        '''
        Imports the script from file filename
        '''
        read_string = self._read_from_file(filename)
        self.set_script(read_string)
        self.compiled = False

    def import_table(self, filename):
        '''
        import the table from filename. ASCII input. tab delimited
        '''
        read_string = self._read_from_file(filename)
        self.parameters.set_ascii_input(read_string)

    def _save_to_file(self, filename, save_string):
        '''
        Save the string to file with filename.
        '''
        try:
            savefile = open(filename, 'w')
        except Exception as e:
            raise GenxIOError(e.__str__(), filename)

        # Save the string to file
        try:
            savefile.write(save_string)
        except Exception as e:
            raise GenxIOError(e.__str__(), filename)

        savefile.close()

    def _read_from_file(self, filename):
        '''
        Reads the entrie file into string and returns it.
        '''
        try:
            loadfile = open(filename, 'r')
        except Exception as e:
            raise GenxIOError(e.__str__(), filename)

        # Read the text from file
        try:
            read_string = loadfile.read()
        except Exception as e:
            raise GenxIOError(e.__str__(), filename)

        loadfile.close()

        return read_string

    def get_parameters(self) -> Parameters:
        return self.parameters

    def get_data(self) -> DataList:
        return self.data

    def get_script(self) -> str:
        return self.script

    def get_filename(self) -> str:
        return self.filename

    def get_possible_parameters(self) -> dict:
        """
        Returns all the parameters that can be fitted. Is used by the parameter grid.
        """
        par_dict = get_parameters(self.script_module, numeric_types_only=True)
        if len(par_dict)==0:
            par_dict = self.get_possible_set_functions()
        return par_dict

    def get_possible_set_functions(self) -> dict:
        """
        Returns all the parameters that can be fitted given by the old style of defining parameters GenX2.4.X
        """
        # Start by updating the config file
        self.ReadConfig()
        # First we should see if any of the 
        # classes is defined in model.__pars__
        # or in __pars__
        pars = []
        try:
            # Check if the have a pars in module named model
            pars_tmp = self.eval_in_model('model.__pars__')
            pars_tmp = ['model.%s'%p for p in pars_tmp]
            pars += pars_tmp
        except:
            pass

        # Check if we have a __pars__ in the main script
        try:
            pars_tmp = self.eval_in_model('__pars__')
            pars_tmp = ['%s'%p for p in pars_tmp]
            pars += pars_tmp
        except:
            pass

        isstrings = sum([type(p)==type('') for p in pars])==len(pars)

        if not isstrings:
            pars = []

        # First find the classes that exists..
        # and defined in self.registred_classes
        classes = []
        for c in self.opt.registred_classes+pars:
            try:
                ctemp = self.eval_in_model(c)
            except:
                pass
            else:
                if inspect.isclass(ctemp):
                    classes.append(ctemp)
        # Check if there are any classes defined before we proceed
        if len(classes)>0:
            # Get all the objects in the compiled module
            names = list(self.script_module.__dict__.keys())
            # Create a tuple of the classes we defined above
            tuple_of_classes = tuple(classes)
            # Creating a dictionary that holds the name of the classes
            # each item for a classes is a new dictionary that holds the
            # object name and then a list of the methods.
            par_dict = {}
            [par_dict.__setitem__(clas.__name__, {}) for clas in classes]
            # find all the names of the objects that belongs to 
            # one of the classes
            objs = [(name, self.eval_in_model(name)) for name in names]
            valid_objs = [(name, obj) for name, obj in objs
                          if isinstance(obj, tuple_of_classes)]
            # nested for loop for finding for each valid object
            # the right name as given by self.model_parameters.set_func
            # Add this to the right item in par_dict given
            # its class and name.
            [par_dict[obj.__class__.__name__].__setitem__(name,
                                                          [member for member in dir(obj)
                                                           if member.startswith(self.opt.set_func)])
             for name, obj in valid_objs]
            return par_dict
        return {}

    def set_script(self, text: str):
        self.script = text
        self.compiled = False

    def set_fom_func(self, fom_func: callable):
        self.fom_func = fom_func

    def is_compiled(self) -> bool:
        return self.compiled

    def bumps_problem(self):
        """
        Return the fit function as a bumps problem to be used with the bumps.fitters.fit function.
        """
        if not self.compiled:
            self.compile_script()
        indep_module = self.script_module
        data = self.data
        # compile again to have a completely independent script module
        self.compile_script()

        from bumps.fitproblem import FitProblem

        return FitProblem(GenxCurve(self))

    def asym_stderr(self, dream_fit):
        """
        Approximate standard error as distance to lower and upper bound of
        the 68% interval for the sample.
        Based on the stderr estimation from dream fit in bumps.
        """
        from bumps.dream.stats import var_stats

        vstats = var_stats(dream_fit.state.draw(portion=dream_fit._trimmed))
        return np.array([(v.p68[0]-v.best, v.p68[1]-v.best) for v in vstats], 'd')

    def bumps_fit(self, method='dream',
                  pop=15, samples=1e5, burn=100, steps=0,
                  thin=1, alpha=0, outliers='none', trim=False,
                  monitors=None, problem=None, **options):
        # create a fitter similar to the bumps.fitters.fit function but with option for GUI monitoring
        if monitors is None:
            monitors = []
        from genx.bumps_optimizer import BumpsResult
        from bumps.fitters import FitDriver, FIT_AVAILABLE_IDS, FITTERS, FIT_ACTIVE_IDS
        options['pop'] = pop
        options['samples'] = samples
        options['burn'] = burn
        options['steps'] = steps
        options['thin'] = thin
        options['alpha'] = alpha
        options['outliers'] = outliers
        options['trim'] = trim

        if problem is None:
            problem = self.bumps_problem()
        problem.fitness.stop_fit = False
        options['abort_test'] = lambda: problem.fitness.stop_fit

        # verbose = True
        if method not in FIT_AVAILABLE_IDS:
            raise ValueError("unknown method %r not one of %s"
                             %(method, ", ".join(sorted(FIT_ACTIVE_IDS))))
        for fitclass in FITTERS:
            if fitclass.id==method:
                break
        # noinspection PyUnboundLocalVariable
        driver = FitDriver(fitclass=fitclass, problem=problem, monitors=monitors, **options)
        driver.clip()  # make sure fit starts within domain
        x0 = problem.getp()
        x, fx = driver.fit()
        problem.setp(x)
        dx = driver.stderr()
        if method=='dream':
            dxpm = self.asym_stderr(driver.fitter)
        else:
            dxpm = None
        result = BumpsResult(x=x, dx=dx, dxpm=dxpm, cov=driver.cov(), chisq=driver.chisq(), bproblem=problem)
        if hasattr(driver.fitter, 'state'):
            result.state = driver.fitter.state
        return result

    def build_dream(self, x_n_gen=25, x_n_chain=5, sigmas=None):
        """
        Compile an independent script model and function to include in the
        Dream MCMC model and return that model.
        """
        if not self.compiled:
            self.compile_script()
        indep_module = self.script_module

        from bumps.dream import Dream
        from bumps.dream.model import Simulation
        from numpy import hstack, random

        data = self.data
        # compile again to have a completely independent script module
        self.compile_script()

        # generate a function that takes a single parameter list and returns the stacked data
        mstring = 'def model_data(p):\n    '
        fstring = ''
        pstart = []
        i = 0
        for p in self.parameters:
            if not p.fit:
                continue
            name = p.name.replace('.set', '_')
            fstring += 'script_module.%s(p[%i])\n'%(p.name, i)
            pstart.append((name, p.value, p.min, p.max))
            i += 1
        fstring += 'res=script_module.Sim(data)\nreturn hstack(res)'
        namespace = {
            'script_module': indep_module, 'data': data,
            'model_data': None, 'hstack': hstack
            }
        exec(mstring+fstring.replace('\n', '\n    '), namespace)

        n = len(pstart)
        if sigmas is None:
            # if no ranges are supplied, use fit range from model
            bounds = (tuple(p[2] for p in pstart), tuple(p[3] for p in pstart))
        else:
            bounds = (tuple(p[1]+3*s[0] for p, s in zip(pstart, sigmas)),
                      tuple(p[1]+3*s[1] for p, s in zip(pstart, sigmas)))
        bsim = Simulation(f=namespace['model_data'],
                          data=hstack([di.y for di in data]),
                          sigma=hstack([di.error for di in data]),
                          bounds=bounds,
                          labels=[p[0] for p in pstart])
        bdream = Dream(model=bsim, draws=20000,
                       population=random.randn(x_n_gen*n, x_n_chain*n, n))  # n_gen, n_chain, n_var
        return bdream

    def bumps_update_parameters(self, res):
        """
        Update the GenX model from a bumps fit result.
        """
        x = res.x
        bproblem = res.bproblem
        names = list(bproblem.labels())
        if len(names)!=len(x):
            raise ValueError('The number of parameters does not fit the model parameters, was the model changed?')

        for pi in self.parameters:
            if pi.fit:
                pi.value = x[names.index(pi.name.replace('.set', '_'))]

    def plot(self, data_labels=None, sim_labels=None):
        """
        Plot all datasets in the model wtih matplotlib similar to the display in the GUI.
        data_labels and sim_labels are lists of strings for each curve,
        if none is supplied the curve label is created from the dataset names and index.
        """
        self.simulate()
        return self.data.plot(data_labels=data_labels, sim_labels=sim_labels)

    def __eq__(self, other: "Model"):
        # compare relevant parts of model object agains another
        return self.data==other.data and self.script==other.script and self.parameters==other.parameters

    def __repr__(self):
        """
        Display information about the model.
        """
        output = "Genx Model"
        if self.compiled:
            output += ' - compiled'
        else:
            output += ' - not compiled yet'
        output += "\n"
        output += "File: %s\n"%self.filename
        output += self.parameters.__repr__()
        output += self.data.__repr__()
        return output

    def _repr_html_(self):
        """
        Display information about the model.
        """
        try:
            self.simulate()
        except Exception as error:
            print(error)

        output = "<h3>Genx Model"
        if self.compiled:
            output += ' - compiled'
        else:
            output += ' - not compiled yet'
        output += "</h3>\n"
        output += "<p>File: %s</p>\n"%self.filename

        output += '<div style="width: 100%;"><div style="width: 40%; float: left;">'
        output += self.data._repr_html_()
        output += "</div>"

        # generate a plot of the model
        import binascii
        from io import BytesIO
        from matplotlib import pyplot as plt
        sio = BytesIO()
        fig = plt.figure(figsize=(10, 8))
        self.data.plot()
        plt.xlabel('q/tth')
        plt.ylabel('Intensity')
        fig.canvas.print_png(sio)
        plt.close()
        img_data = binascii.b2a_base64(sio.getvalue()).decode('ascii')
        output += '<div style="margin-left: 40%;"><img src="data:image/png;base64,{}&#10;"></div></div>'.format(
            img_data)

        output += self.parameters._repr_html_()
        return output

    @property
    def widget(self):
        return self._repr_ipyw_()

    def _repr_ipyw_(self):
        import ipywidgets as ipw
        try:
            self.simulate()
        except Exception as error:
            print(error)

        graphw = ipw.Output()
        with graphw:
            from matplotlib import pyplot as plt
            fig = plt.figure(figsize=(10, 8))
            self.data.plot()
            plt.xlabel('q/tth')
            plt.ylabel('Intensity')
            from IPython.display import display
            display(fig)
            plt.close()

        dataw = self.data._repr_ipyw_()
        top = ipw.HBox([dataw, graphw])
        parameters = self.parameters._repr_ipyw_()

        replot = ipw.Button(description='simulate')
        replot.on_click(self._ipyw_replot)
        replot._plot_output = graphw

        script_area = ipw.Textarea(self.script, layout=ipw.Layout(width='100%'))
        script_area.observe(self._ipyw_script, names='value')
        tabs = ipw.Tab(children=[parameters, script_area])
        tabs.set_title(0, 'Parameters')
        tabs.set_title(1, 'Script')

        return ipw.VBox([replot, top, tabs])

    def _ipyw_replot(self, button):
        try:
            self.simulate()
        except Exception as error:
            print(error)
        with button._plot_output:
            from IPython.display import display, clear_output
            from matplotlib import pyplot as plt
            fig = plt.figure(figsize=(10, 8))
            self.data.plot()
            plt.xlabel('q/tth')
            plt.ylabel('Intensity')
            plt.draw()
            clear_output()
            display(fig)
            plt.close()

    def _ipyw_script(self, change):
        self.set_script(change.new)


class GenxCurve:
    """
    Bumps Curve object for a GenX model.
    """
    model: Model
    n_fev: int

    def __init__(self, model: Model):
        self._num_curves = 1

        self.labels = ['x', 'I']

        self.model = model

        if not self.model.compiled:
            self.model.compile_script()
        self.model_script = self.model.script_module

        self.name = "GenX model"
        self.plot_x = None

        pars, state, funcs = self._parse_pars()

        self._pnames = list(pars.keys())
        self._pars = pars
        self._state = state
        self._set_funcs = funcs
        self._cached_theory = None
        self.stop_fit = False
        self.n_fev = 0

    @property
    def x(self):
        return [self._pars[name].value for name in self._pnames]

    @property
    def y(self):
        return np.hstack([di.y for di in self.model.data if di.use])

    @property
    def dy(self):
        return np.hstack([di.error for di in self.model.data if di.use])

    def _parse_pars(self):
        from bumps.parameter import Parameter
        pars = {}
        state = {}
        funcs = {}
        for p in self.model.parameters:
            if p.name.strip()=='':
                continue
            name = p.name.replace('.set', '_')
            if p.fit:
                pars[name] = Parameter.default(p.value, name=name, bounds=(p.min, p.max))
            else:
                state[name] = p.value
            funcs[name] = self.model.create_fit_func(p.name)
        return pars, state, funcs

    def update(self):
        self._cached_theory = None
        self.n_fev = 0

    def parameters(self):
        return self._pars

    def numpoints(self):
        return np.sum([di.x.shape[0] for di in self.model.data if di.use])

    def theory(self, x=None):
        # Use cache if x is None, otherwise compute theory with x.
        if x is None:
            if self._cached_theory is None:
                self._cached_theory = self._compute_theory(self.x)
            return self._cached_theory
        return self._compute_theory(x)

    def _compute_theory(self, x):
        self._apply_par(x)
        sim = self.model_script.Sim(self.model.data)
        self.n_fev += 1
        return np.hstack([si for si, di in zip(sim, self.model.data) if di.use])

    def _apply_par(self, x):
        for ni, si in self._state.items():
            self._set_funcs[ni](si)
        for ni, xi in zip(self._pnames, x):
            self._set_funcs[ni](xi)

    def simulate_data(self, noise=None):
        theory = self.theory()
        if noise is not None:
            if noise=='data':
                pass
            elif noise<0:
                self.dy = -0.01*noise*theory
            else:
                self.dy = noise
        self.y = theory+np.random.randn(*theory.shape)*self.dy

    def residuals(self):
        return (self.theory()-self.y)/self.dy

    def nllf(self):
        r = self.residuals()
        fom = np.sum(r**2)
        penalty_funcs = self.model.get_par_penalty()
        if len(penalty_funcs)>0 and fom is not np.NAN:
            fom += sum([pf() for pf in penalty_funcs])*(len(r)-len(self._pars))
        return 0.5*fom

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['model_script']
        del state['_set_funcs']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        if not self.model.compiled:
            self.model.compile_script()
        self.model_script = self.model.script_module

        pars, state, funcs = self._parse_pars()
        self._pnames = list(pars.keys())
        self._pars = pars
        self._state = state
        self._set_funcs = funcs
